# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2017-2018
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This module contains and manages kas plugins
"""

PLUGINS = {}


def register_plugins(mod):
    """
        Register all kas plugins found in a module
    """
    for plugin in getattr(mod, '__KAS_PLUGINS__', []):
        PLUGINS[plugin.name] = plugin


def load():
    """
        Import all kas plugins
    """
    # sort alphabetically
    from . import build
    from . import checkout
    from . import clean
    from . import diff
    from . import dump
    from . import for_all_repos
    from . import lock
    from . import menu
    from . import shell

    register_plugins(build)
    register_plugins(checkout)
    register_plugins(clean)
    register_plugins(diff)
    register_plugins(dump)
    register_plugins(for_all_repos)
    register_plugins(lock)
    register_plugins(menu)
    register_plugins(shell)


def get(name):
    """
        Lookup a kas plugin class by name
    """
    return PLUGINS.get(name, None)


def all():
    """
        Get a list of all loaded kas plugin classes
    """
    return PLUGINS.values()



# kas - setup tool for bitbake based projects
#
# Copyright (c) Konsulko Group, 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This plugin implements the ``kas checkout`` command.

    When this command is executed, kas will checkout repositories and set up
    the build directory as specified in the chosen config file. This command
    is useful if you need to inspect the configuration or modify any of the
    checked out layers before starting a build.

    For example, to setup the configuration described in the file
    ``kas-project.yml`` you could run::

        kas checkout kas-project.yml
"""

from kas.context import create_global_context
from kas.config import Config
from kas.libcmds import Macro
from kas.libkas import setup_parser_common_args, setup_parser_config_arg

__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens AG, 2017-2018'


class Checkout:
    name = 'checkout'
    helpmsg = (
        'Checks out all necessary repositories and sets up the build '
        'directory as specified in the configuration file.'
    )

    @classmethod
    def setup_parser(cls, parser):
        setup_parser_common_args(parser)
        setup_parser_config_arg(parser)

    def run(self, args):
        ctx = create_global_context(args)
        ctx.config = Config(ctx, args.config)

        macro = Macro()
        macro.run(ctx, args.skip)


__KAS_PLUGINS__ = [Checkout]



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2019
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import stat
import shutil
import pytest
import subprocess
from kas import kas
from kas.repos import PatchApplyError, PatchFileNotFound, PatchMappingError


def git_get_commit(path):
    output = subprocess.check_output(
        ['git', 'rev-parse', 'HEAD'], cwd=path)
    return output.decode('utf-8').strip()


def mercurial_get_commit(path):
    output = subprocess.check_output(
        ['hg', 'log', '-r', '.', '--template', '{node}\n'], cwd=path)
    return output.decode('utf-8').strip()


@pytest.mark.online
def test_patch(monkeykas, tmpdir, mercurial):
    tdir = tmpdir / 'test_patch'
    shutil.copytree('tests/test_patch', tdir)
    monkeykas.chdir(tdir)
    kas_wd = monkeykas.get_kwd()

    repo = mercurial(tmpdir, 'example')
    commit = repo.get_commit()
    subprocess.check_call(
        ['sed', '-i', f's/82e55d328c8c/{commit}/', 'test.yml'])

    kas.kas(['shell', 'test.yml', '-c', 'true'])
    for f in [kas_wd / 'kas/tests/test_patch/hello.sh',
              kas_wd / 'hello/hello.sh']:
        assert os.stat(f)[stat.ST_MODE] & stat.S_IXUSR

    kas_head_ref = git_get_commit(kas_wd / "kas")
    kas_branch_head_ref = git_get_commit(kas_wd / "kas-branch")
    hello_head_ref = mercurial_get_commit(kas_wd / "hello")
    hello_branch_head_ref = mercurial_get_commit(kas_wd / "hello-branch")

    kas.kas(['shell', 'test.yml', '-c', 'true'])

    assert git_get_commit(kas_wd / "kas") == kas_head_ref
    assert git_get_commit(kas_wd / "kas-branch") == kas_branch_head_ref
    assert mercurial_get_commit(kas_wd / "hello") == hello_head_ref
    assert mercurial_get_commit(kas_wd / "hello-branch") == \
        hello_branch_head_ref


@pytest.mark.online
def test_patch_update(monkeykas, tmpdir, mercurial):
    """
        Test that patches are applied correctly after switching a repo from
        a branch to a commit hash and vice-versa with both git and mercurial
        repositories.
    """
    tdir = str(tmpdir / 'test_patch_update')
    shutil.copytree('tests/test_patch', tdir)
    monkeykas.chdir(tdir)
    kas_wd = monkeykas.get_kwd()

    repo = mercurial(tmpdir, 'example')
    commit = repo.get_commit()
    for file, c in [('test.yml', '82e55d328c8c'),
                    ('test2.yml', '0a04b987be5a')]:
        subprocess.check_call(
            ['sed', '-i', f's/{c}/{commit}/', file])

    kas.kas(['shell', 'test.yml', '-c', 'true'])
    kas.kas(['shell', 'test2.yml', '-c', 'true'])
    for f in [kas_wd / 'kas/tests/test_patch/hello.sh',
              kas_wd / 'hello/hello.sh']:
        assert os.stat(f)[stat.ST_MODE] & stat.S_IXUSR


@pytest.mark.online
def test_invalid_patch(monkeykas, tmpdir):
    """
        Test on common errors when applying patches
    """
    tdir = str(tmpdir / 'test_patch_invalid')
    shutil.copytree('tests/test_patch', tdir)
    monkeykas.chdir(tdir)

    with pytest.raises(PatchFileNotFound):
        kas.kas(['shell', 'test-invalid.yml', '-c', 'true'])

    with pytest.raises(PatchMappingError):
        kas.kas(['shell', 'test-invalid2.yml', '-c', 'true'])

    with pytest.raises(PatchApplyError):
        kas.kas(['shell', 'test-invalid3.yml', '-c', 'true'])


@pytest.mark.online
def test_patch_dirty_repo(monkeykas, tmpdir, mercurial):
    """
        Test that kas will not apply patches to a dirty repository
    """
    tdir = str(tmpdir / 'test_patch_dirty')
    shutil.copytree('tests/test_patch', tdir)
    monkeykas.chdir(tdir)

    repo = mercurial(tmpdir, 'example')
    commit = repo.get_commit()
    subprocess.check_call(
        ['sed', '-i', f's/82e55d328c8c/{commit}/', 'test.yml'])

    kas.kas(['checkout', '--skip', 'repos_apply_patches', 'test.yml'])

    kas_branch_dir = monkeykas.get_kwd() / 'kas-branch'
    with open(kas_branch_dir / 'README.rst', 'a') as f:
        f.write('echo "dirty"')
    kas.kas(['checkout', 'test.yml'])
    # the dirty repo must not be patched
    assert not (kas_branch_dir / 'tests/test_patch/hello.sh').exists()
    # the clean repo must be patched
    assert (monkeykas.get_kwd() / 'hello-branch/hello.sh').exists()



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2017-2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This module contains the implementation of the kas configuration.
"""

import os
import json
import copy
from pathlib import Path
from .repos import Repo
from .includehandler import IncludeHandler
from .kasusererror import ArtifactNotFoundError
from .configschema import CONFIGSCHEMA

__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens AG, 2017-2021'

CONFIG_YAML_FILE = '.config.yaml'


class Config:
    """
        Implements the kas configuration based on config files.
    """
    def __init__(self, ctx, filename, target=None, task=None):
        self._override_target = target
        self._override_task = task
        self._build_dir = ctx.build_dir
        self.__config = {}
        if not filename:
            filename = os.path.join(ctx.kas_work_dir, CONFIG_YAML_FILE)

        self.filenames = [os.path.abspath(configfile)
                          for configfile in filename.split(':')]

        update = ctx.args.update if hasattr(ctx.args, 'update') else False

        self.handler = IncludeHandler(self.filenames, not update)
        self.repo_dict = {}
        self.repo_cfg_hashes = {}

    @property
    def _config(self):
        if not self.__config:
            raise RuntimeError("Config has not been imported yet.")
        return self.__config

    def get_build_system(self):
        """
            Returns the pre-selected build system
        """
        return self._config.get('build_system', '')

    def find_missing_repos(self, repo_paths={}):
        """
            Returns repos that are in config but not on disk and updates
            the internal config dictionary.
        """
        (self.__config, missing_repo_names) = \
            self.handler.get_config(repos=repo_paths)

        return missing_repo_names

    def get_config(self, remove_includes=False, apply_overrides=False):
        """
            Returns a copy of the config dict
        """
        config = copy.deepcopy(self._config)
        if remove_includes and 'includes' in config['header']:
            del config['header']['includes']
        if apply_overrides:
            overrides = config.get('overrides', {})
            if overrides:
                config.update(overrides)
                del config['overrides']
        return config

    def get_lockfiles(self):
        """
            Returns the list of in-use lockfiles.
        """
        return self.handler.get_lockfiles()

    def get_repos_config(self):
        """
            Returns the repository configuration
        """
        return self._config.get('repos', {})

    def get_repos(self):
        """
            Returns the list of repos.
        """

        # Always keep repo_dict and repos synchronous
        # when calling get_repos
        self.repo_dict = self._get_repo_dict()
        return list(self.repo_dict.values())

    def get_repo(self, name):
        """
            Returns a `Repo` instance for the configuration with the key
            `name`.
        """
        repo_defaults = self._config.get('defaults', {}).get('repos', {})
        overrides = self._config.get('overrides', {}) \
                                .get('repos', {}).get(name, {})
        config = self.get_repos_config()[name] or {}
        top_repo_path = self.handler.get_top_repo_path()

        # Check if we have this repo with an identical config already.
        # As this function is called across various places and with different
        # configurations (e.g. due to updates from transitive includes),
        # we cache the results.
        args = (name, config, repo_defaults, top_repo_path, overrides)
        return self._get_or_create_repo(args)

    def _get_or_create_repo(self, args):
        """
            Get a repo from the cache and insert it if not existing.
            Creating repos is expensive due to external commands being called.
        """
        encoded = json.dumps(args, sort_keys=True).encode()
        if encoded in self.repo_cfg_hashes:
            return self.repo_cfg_hashes[encoded]
        repo = Repo.factory(*args)
        self.repo_cfg_hashes[encoded] = repo
        return repo

    def _get_repo_dict(self):
        """
            Returns a dictionary containing the repositories with
            their names (as it is defined in the config file) as keys
            and the `Repo` instances as values.
        """
        return {name: self.get_repo(name) for name in self.get_repos_config()}

    def get_bitbake_targets(self):
        """
            Returns a list of bitbake targets
        """
        if self._override_target:
            return self._override_target
        environ_targets = [i
                           for i in os.environ.get('KAS_TARGET', '').split()
                           if i]
        if environ_targets:
            return environ_targets
        def_target = CONFIGSCHEMA['properties']['target']['default']
        target = self._config.get('target', def_target)
        if isinstance(target, str):
            return [target]
        return target

    def get_bitbake_task(self):
        """
            Returns the bitbake task
        """
        if self._override_task:
            return self._override_task
        default = CONFIGSCHEMA['properties']['task']['default']
        return os.environ.get('KAS_TASK',
                              self._config.get('task', default))

    def _get_conf_header(self, header_name):
        """
            Returns the local.conf header
        """
        header = ''
        for key, value in sorted(self._config.get(header_name, {}).items()):
            header += f'# {key}\n{value}\n'
        return header

    def get_bblayers_conf_header(self):
        """
            Returns the bblayers.conf header
        """
        return self._get_conf_header('bblayers_conf_header')

    def get_local_conf_header(self):
        """
            Returns the local.conf header
        """
        return self._get_conf_header('local_conf_header')

    def get_machine(self):
        """
            Returns the machine
        """
        default = CONFIGSCHEMA['properties']['machine']['default']
        return os.environ.get('KAS_MACHINE',
                              self._config.get('machine', default))

    def get_distro(self):
        """
            Returns the distro
        """
        default = CONFIGSCHEMA['properties']['distro']['default']
        return os.environ.get('KAS_DISTRO',
                              self._config.get('distro', default))

    def get_environment(self):
        """
            Returns the configured environment variables from the configuration
            file with possible overwritten values from the environment.
        """
        env = self._config.get('env', {})
        return {var: os.environ.get(var, env[var]) for var in env}

    def get_multiconfig(self):
        """
            Returns the multiconfig array as bitbake string
        """
        multiconfigs = set()
        for target in self.get_bitbake_targets():
            if target.startswith('multiconfig:') or target.startswith('mc:'):
                multiconfigs.add(target.split(':')[1])
        return ' '.join(multiconfigs)

    def get_artifacts(self, missing_ok=True):
        """
            Returns the found artifacts after glob expansion, relative
            to the build_dir as a list of tuples (name, path).
            If missing_ok=False, raises an ArtifactNotFoundError if no
            artifact for a given name is found.
        """
        arts = self._config.get('artifacts', {})
        foundfiles = []
        for name, art in arts.items():
            files = list(Path(self._build_dir).glob(art))
            if not missing_ok and len(files) == 0:
                raise ArtifactNotFoundError(name, art)
            foundfiles.extend([(name, f) for f in files])
        return [(n, f.relative_to(self._build_dir))
                for n, f in foundfiles]

    def get_signers_config(self, keytype=None):
        """
            Returns the keys from the configuration
        """
        signers = self._config.get('signers', {})
        if not keytype:
            return signers
        else:
            return {k: v for k, v in signers.items()
                    if v.get('type', 'gpg') == keytype}

    def get_buildtools(self):
        """
            Returns the buildtools keys: version, download URL and
            archive filename. These are provided so kas knows which
            buildtools archive to fetch and from what source.
        """

        return self._config.get('buildtools')



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2019-2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import signal
import pytest
import os
import subprocess
import shutil
from pathlib import Path

from kas import kas

ENVVARS_KAS = [
    'KAS_WORK_DIR',
    'KAS_BUILD_DIR',
    'KAS_REPO_REF_DIR',
    'KAS_DISTRO',
    'KAS_MACHINE',
    'KAS_TARGET',
    'KAS_TASK',
    'KAS_PREMIRRORS',
    'KAS_CLONE_DEPTH',
    'KAS_CONTAINER_SCRIPT_VERSION',
    'SSH_PRIVATE_KEY',
    'SSH_PRIVATE_KEY_FILE',
    'SSH_AUTH_SOCK',
    'CI_SERVER_HOST',
    'CI_JOB_TOKEN',
    'CI_PROJECT_DIR',
    'GITLAB_CI',
    'GITHUB_ACTIONS',
    'REMOTE_CONTAINERS'
]

ENVVARS_TOOLS = [
    'EMAIL'
]


class KasEnvConfig:
    def __init__(self, work_dir: str, build_dir: str):
        self.work_dir = work_dir
        self.build_dir = build_dir


KAS_OPERATING_DIRS = [
    KasEnvConfig(None, None),
    KasEnvConfig("kas-work", None),
    KasEnvConfig("kas-work", "kas-build"),
]
KAS_OPERATING_IDS = ["noenv", "work", "workbuild"]


# only fan-out offline tests
def pytest_generate_tests(metafunc):
    if "monkeykas" not in metafunc.fixturenames:
        return
    if next(metafunc.definition.iter_markers("dirsfromenv"), None):
        metafunc.parametrize("monkeykas",
                             KAS_OPERATING_DIRS,
                             ids=KAS_OPERATING_IDS,
                             indirect=True)
    else:
        metafunc.parametrize("monkeykas",
                             [KAS_OPERATING_DIRS[0]],
                             ids=[KAS_OPERATING_IDS[0]],
                             indirect=True)


@pytest.fixture(autouse=True)
def prepare_kas_for_tests(monkeypatch):
    """
        Patch-out some global logic kas modifies. The strategy is as
        following: kas implements the problematic parts as dedicated
        functions which we monkeypatch here.
    """
    def _register_signal_handlers(loop):
        loop.add_signal_handler(signal.SIGTERM, kas.interruption)
        # do not handle signal.SIGINT

    def _set_global_loglevel(level):
        # do not set global log level as this is handled by pytest
        pass

    monkeypatch.setattr(kas, "register_signal_handlers",
                        _register_signal_handlers)
    monkeypatch.setattr(kas, "set_global_loglevel", _set_global_loglevel)


@pytest.fixture()
def monkeykas(request, monkeypatch, tmpdir):
    def get_kas_work_dir():
        work_dir = os.environ.get('KAS_WORK_DIR', '.')
        return Path(work_dir).absolute()

    def get_kas_build_dir():
        build_dir = os.environ.get('KAS_BUILD_DIR')
        if (not build_dir):
            return get_kas_work_dir() / 'build'
        return Path(build_dir).absolute()

    def move_to_workdir(orig_path):
        kas_wd = get_kas_work_dir()
        _orig = Path(orig_path)
        if not _orig.parent.samefile(kas_wd):
            shutil.move(_orig, kas_wd / _orig.basename())

    monkeypatch.get_kwd = get_kas_work_dir
    monkeypatch.get_kbd = get_kas_build_dir
    monkeypatch.move_to_kwd = move_to_workdir

    _work_dir = request.param.work_dir
    if _work_dir:
        workdir = tmpdir / _work_dir
        workdir.mkdir()
        monkeypatch.setenv('KAS_WORK_DIR', str(workdir))
    _build_dir = request.param.build_dir
    if _build_dir:
        builddir = workdir / _build_dir
        builddir.mkdir()
        monkeypatch.setenv('KAS_BUILD_DIR', str(builddir))

    for var in ENVVARS_KAS + ENVVARS_TOOLS:
        monkeypatch.delenv(var, raising=False)
    # Set HOME to a temporary directory
    homedir = tmpdir / '_home'
    homedir.mkdir()
    monkeypatch.setenv('HOME', str(homedir))

    # remove all VSCode devcontainers related variables
    for var in os.environ.keys():
        if var.startswith('REMOTE_CONTAINERS_'):
            monkeypatch.delenv(var)
    # remove all git related variables
    for var in os.environ.keys():
        if var.startswith('GIT_'):
            monkeypatch.delenv(var)
    # provide minimal git environment
    monkeypatch.setenv('GIT_AUTHOR_NAME', 'kas')
    monkeypatch.setenv('GIT_AUTHOR_EMAIL', 'kas@example.com')
    monkeypatch.setenv('GIT_COMMITTER_NAME', 'kas')
    monkeypatch.setenv('GIT_COMMITTER_EMAIL', 'kas@example.com')

    yield monkeypatch


class MercurialRepo:
    """
    Create a new Mercurial repository with a single commit.
    """
    def __init__(self, tmpdir, name, branch=None):
        """
        Creates a new Mercurial repository with a single commit.
        The content resembles the Makefile from the hello world example.
        """
        self.repo = tmpdir / name
        self.repo.mkdir()
        subprocess.check_call(['hg', 'init'], cwd=self.repo)
        if branch:
            subprocess.check_call(['hg', 'branch', branch], cwd=self.repo)
        with open(self.repo / 'Makefile', 'w') as f:
            f.write('all:\n\techo hello\n')
        subprocess.check_call(['hg', 'add', 'Makefile'], cwd=self.repo)
        subprocess.check_call(['hg', '--config', 'ui.username=kas',
                               'commit', '-m', 'initial commit',
                               '--date', '2024-11-10 08:00'], cwd=self.repo)

    def __enter__(self):
        return self.repo

    def get_commit(self):
        return subprocess.check_output(
            ['hg', 'log', '-r', '.', '--template', '{rev}:{node}'],
            cwd=self.repo).decode('utf-8').strip()

    def __exit__(self, exc_type, exc_value, traceback):
        self.repo.remove()


@pytest.fixture
def mercurial():
    def make_hg_repo(tmpdir, name, branch=None):
        return MercurialRepo(tmpdir, name, branch)
    return make_hg_repo



# kas - setup tool for bitbake based projects
#
# Copyright (c) Konsulko Group, 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import glob
import os
import pathlib
import shutil
import json
import yaml
import subprocess
import pytest
from kas import kas
from kas.libkas import TaskExecError, KasUserError, run_cmd
from kas.attestation import file_digest_slow
from kas.repos import RepoFetchError, RepoRefError


@pytest.mark.dirsfromenv
@pytest.mark.online
def test_for_all_repos(monkeykas, tmpdir):
    tdir = str(tmpdir / 'test_commands')
    shutil.copytree('tests/test_commands', tdir)
    monkeykas.chdir(tdir)
    kas.kas(['for-all-repos', 'test.yml',
             '''if [ -n "${KAS_REPO_URL}" ]; then git rev-parse HEAD \
                     >> %s/ref_${KAS_REPO_NAME}; fi''' % tdir])

    with open('ref_kas_1.0', 'r') as f:
        assert f.readline().strip() \
            == '907816a5c4094b59a36aec12226e71c461c05b77'
    with open('ref_kas_1.1', 'r') as f:
        assert f.readline().strip() \
            == 'e9ca55a239caa1a2098e1d48773a29ea53c6cab2'


@pytest.mark.online
def test_for_all_repos_keep_config_unchanged(monkeykas, tmpdir):
    tdir = str(tmpdir / 'test_commands')
    shutil.copytree('tests/test_commands', tdir)
    monkeykas.chdir(tdir)

    with pytest.raises(FileNotFoundError, match=r'.*/kas_1.[01]'):
        kas.kas(['for-all-repos', '--keep-config-unchanged', 'test.yml',
                 'pwd'])

    assert not os.path.exists('kas_1.0')
    assert not os.path.exists("kas_1.1")


@pytest.mark.online
def test_checkout(monkeykas, tmpdir):
    tdir = str(tmpdir / 'test_commands')
    shutil.copytree('tests/test_commands', tdir)
    monkeykas.chdir(tdir)
    kas_bd = monkeykas.get_kbd()
    kas.kas(['checkout', 'test.yml'])

    # Ensure that local.conf and bblayers.conf are populated, check that no
    # build has been executed by ensuring that no tmp, sstate-cache or
    # downloads directories are present.
    assert (kas_bd / 'conf/local.conf').exists()
    assert (kas_bd / 'conf/bblayers.conf').exists()
    assert not glob.glob(f'{str(kas_bd)}/tmp*')
    assert not (kas_bd / 'downloads').exists()
    assert not (kas_bd / 'sstate-cache').exists()

    # test re-fetching (remove commit as nothing is fetched otherwise)
    with open('test.yml', 'r') as f:
        yml = yaml.safe_load(f)
    with open('test-no-commit.yml', 'w') as f:
        del yml['repos']['kas_1.1']['commit']
        yml['repos']['kas_1.1']['branch'] = 'master'
        yaml.safe_dump(yml, f)
    kas.kas(['checkout', '--update', 'test-no-commit.yml'])

    with monkeykas.context() as mc:
        from kas.repos import GitRepo
        mc.setattr(GitRepo, "fetch_cmd", lambda _: ["false"])
        with pytest.raises(RepoFetchError):
            kas.kas(['checkout', '--update', 'test-no-commit.yml'])


@pytest.mark.parametrize(
    "file, error",
    [
        ('test-invalid.yml', TaskExecError),
        ('test-invalid-tag.yml', RepoRefError),
    ],
)
@pytest.mark.online
def test_invalid_checkout(monkeykas, tmpdir, capsys, file, error):
    tdir = str(tmpdir / 'test_commands')
    shutil.copytree('tests/test_commands', tdir)
    monkeykas.chdir(tdir)
    with pytest.raises(error):
        kas.kas(['checkout', file])


@pytest.mark.dirsfromenv
@pytest.mark.online
def test_checkout_with_ci_rewrite(monkeykas, tmpdir):
    tdir = str(tmpdir / 'test_commands')
    shutil.copytree('tests/test_commands', tdir)
    monkeykas.chdir(tdir)
    project_dir = str(tmpdir.mkdir('build'))
    with monkeykas.context() as mp:
        mp.setenv('GITLAB_CI', 'true')
        mp.setenv('CI_SERVER_HOST', 'github.com')
        mp.setenv('CI_JOB_TOKEN', 'not-needed')
        mp.setenv('CI_PROJECT_DIR', project_dir)
        kas.kas(['checkout', 'test-url-rewrite.yml'])


@pytest.mark.online
def test_checkout_create_refs(monkeykas, tmpdir):
    tdir = str(tmpdir / 'test_commands')
    repo_cache = pathlib.Path(str(tmpdir.mkdir('repos')))
    shutil.copytree('tests/test_commands', tdir)
    monkeykas.chdir(tdir)
    monkeykas.setenv('KAS_REPO_REF_DIR', str(repo_cache))
    kas_wd = monkeykas.get_kwd()
    kas.kas(['checkout', 'test.yml'])
    assert (repo_cache / 'github.com.siemens.kas.git').exists()
    assert (kas_wd / 'kas_1.0/.git/objects/info/alternates').exists()
    # check if refs are removed on purge
    kas.kas(['purge', 'test.yml'])
    assert not (repo_cache / 'github.com.siemens.kas.git').exists()


@pytest.mark.online
def test_checkout_shallow(monkeykas, tmpdir):
    tdir = str(tmpdir / 'test_commands')
    shutil.copytree('tests/test_commands', tdir)
    monkeykas.chdir(tdir)
    with monkeykas.context() as mp:
        mp.setenv('KAS_CLONE_DEPTH', 'invalid')
        with pytest.raises(KasUserError):
            kas.kas(['checkout', 'test-shallow.yml'])

    with monkeykas.context() as mp:
        mp.setenv('KAS_CLONE_DEPTH', '1')
        kas.kas(['checkout', 'test-shallow.yml'])
    for repo in ['kas_1', 'kas_2', 'kas_3', 'kas_4', 'kas_5']:
        repo_path = monkeykas.get_kwd() / repo
        output = subprocess.check_output(
            ['git', 'rev-list', '--count', 'HEAD'], cwd=repo_path)
        count = int(output.decode('utf-8').strip())
        if repo == 'kas_4':
            assert count >= 1
        else:
            assert count == 1


@pytest.mark.online
def test_shallow_updates(monkeykas, tmpdir):
    def _get_commit(repo):
        output = subprocess.check_output(
            ['git', 'rev-parse', '--verify', 'HEAD'], cwd=repo)
        return output.decode('utf-8').strip()

    tdir = tmpdir / 'test_commands'
    tdir.mkdir()
    shutil.copy('tests/test_commands/oe-init-build-env', tdir)
    monkeykas.chdir(tdir)
    monkeykas.setenv('KAS_CLONE_DEPTH', '1')
    kas_wd = monkeykas.get_kwd()
    # test non-pinned checkout of master branch
    base_yml = {'header': {'version': 15}, 'repos': {
                'this': {},
                'kas': {
                    'url': 'https://github.com/siemens/kas.git',
                    'branch': 'master'
                }}}
    with open(tdir / 'kas.yml', 'w') as f:
        yaml.dump(base_yml, f)
    kas.kas(['checkout', 'kas.yml'])
    # switch branches, perform checkout again
    base_yml['repos']['kas']['branch'] = 'next'
    with open(tdir / 'kas.yml', 'w') as f:
        yaml.dump(base_yml, f)
    kas.kas(['checkout', 'kas.yml'])
    # pin commit on next branch
    commit = '5d1ab6e8ed3a12c7093c9041f104fb6a2db701a1'
    base_yml_lock = {'header': {'version': 15},
                     'overrides': {'repos': {'kas': {'commit': commit}}}}
    with open(tdir / 'kas.lock.yml', 'w') as f:
        yaml.dump(base_yml_lock, f)
    kas.kas(['checkout', 'kas.yml'])
    assert _get_commit(kas_wd / 'kas') == commit
    # update to latest revision of next branch
    kas.kas(['checkout', '--update', 'kas.yml'])
    assert _get_commit(kas_wd / 'kas') != commit


@pytest.mark.dirsfromenv
@pytest.mark.online
def test_repo_includes(monkeykas, tmpdir):
    tdir = tmpdir / 'test_commands'
    shutil.copytree('tests/test_repo_includes', tdir)
    monkeykas.chdir(tdir)
    monkeykas.move_to_kwd('subrepo')
    kas.kas(['checkout', 'test.yml'])


@pytest.mark.online
def test_dump(monkeykas, tmpdir, capsys):
    tdir = tmpdir / 'test_commands'
    shutil.copytree('tests/test_repo_includes', tdir)
    monkeykas.chdir(tdir)
    kas_bd = monkeykas.get_kbd()
    monkeykas.move_to_kwd('subrepo')

    formats = ['json', 'yaml']
    resolve = ['', '--resolve-refs', '--resolve-env']
    # test cross-product of these options (formats x resolve)
    for f, r in ((f, r) for f in formats for r in resolve):
        outfile = 'test_flat%s.%s' % (r, f)

        with monkeykas.context() as mp:
            if r == '--resolve-env':
                mp.setenv('TESTVAR_FOO', 'KAS')
            kas.kas(('dump --format %s %s test.yml' % (f, r)).split())

        with open(outfile, 'w') as file:
            file.write(capsys.readouterr().out)

        with open(outfile, 'r') as cf:
            flatconf = json.load(cf) if f == 'json' else yaml.safe_load(cf)
            commit = flatconf['repos']['kas3'].get('commit', None)
            envvar = flatconf['env']['TESTVAR_FOO']
            if r == '--resolve-refs':
                assert commit is not None
            else:
                assert commit is None
            if r == '--resolve-env':
                assert envvar == 'KAS'
            else:
                assert envvar == 'BAR'

            assert 'includes' not in flatconf['header']
            # check if kas can read the generated file
            if f == 'yaml':
                shutil.rmtree(kas_bd, ignore_errors=True)
                kas.kas(('checkout %s' % outfile).split())
                assert (kas_bd / 'conf/local.conf').exists()


@pytest.mark.dirsfromenv
@pytest.mark.online
def test_diff(monkeykas, tmpdir, capsys):
    tdir = str(tmpdir / 'test_commands')
    shutil.copytree('tests/test_diff', tdir)
    monkeykas.chdir(tdir)

    diff = kas.kas(('diff diff1_folder/diff1.yml diff2_folder/diff2.yml')
                   .split())
    assert len(capsys.readouterr().out) > 0

    diff = kas.kas(('diff diff4.yml:diff5.yml diff3.yml').split())
    assert len(capsys.readouterr().out) > 0

    diff = kas.kas(('diff diff3_cus_repopath.yml diff3.yml').split())
    assert len(capsys.readouterr().out) > 0

    diff = kas.kas(('diff  --format json diff1_folder/diff1.yml '
                    'diff2_folder/diff2.yml')
                   .split())
    outfile = 'diff-output.json'
    with open(outfile, 'w') as file:
        file.write(capsys.readouterr().out)
    with open(outfile, 'r') as cf:
        diff = json.load(cf)
        assert diff is not None
        assert diff["values_changed"] is not None
        assert diff["values_changed"]["target"]["old_value"] == "target1"
        assert diff["values_changed"]["target"]["new_value"] == "target2"
        assert diff["values_changed"]["machine"]["old_value"] == "machine1"
        assert diff["values_changed"]["machine"]["new_value"] == "machine2"
        assert diff["values_changed"]["distro"]["old_value"] == "distro1"
        assert diff["values_changed"]["distro"]["new_value"] == "distro2"
        assert len(diff["vcs"]["isar"]) == 59


@pytest.mark.dirsfromenv
@pytest.mark.online
def test_lockfile(monkeykas, tmpdir, capsys):
    tdir = tmpdir.mkdir('test_commands')
    shutil.rmtree(tdir, ignore_errors=True)
    shutil.copytree('tests/test_repo_includes', tdir)
    monkeykas.chdir(tdir)
    kas_wd = monkeykas.get_kwd()
    monkeykas.move_to_kwd('subrepo')

    # no lockfile yet, branches are floating
    kas.kas('dump test.yml'.split())
    rawspec = yaml.safe_load(capsys.readouterr().out)
    assert rawspec['repos']['externalrepo']['refspec'] == 'master'

    with open(kas_wd / 'externalrepo/.git/refs/heads/master') as f:
        expected_commit = f.readline().strip()

    # create lockfile
    kas.kas('dump --lock --inplace test.yml'.split())
    assert os.path.exists('test.lock.yml')

    # check if legacy dump -> lock redirection works
    with open('test.lock.yml', "rb") as f:
        hash_dump = file_digest_slow(f, 'sha256')
    os.remove('test.lock.yml')
    kas.kas('lock test.yml'.split())
    with open('test.lock.yml', "rb") as f:
        hash_lock = file_digest_slow(f, 'sha256')
    assert hash_dump.hexdigest() == hash_lock.hexdigest()

    # lockfile is considered during import, expect pinned branches
    kas.kas('dump test.yml'.split())
    lockspec = yaml.safe_load(capsys.readouterr().out)
    assert lockspec['overrides']['repos']['externalrepo']['commit'] \
        == expected_commit

    # insert older commit into lockfile (kas post commit/branch introduction)
    test_commit = '226e92a7f30667326a63fd9812b8cc4a6184e398'
    # a commit message between test_commit and HEAD
    test_shortmsg = 'Release 4.8'
    lockspec['overrides']['repos']['externalrepo']['commit'] = test_commit
    with open('test.lock.yml', 'w') as f:
        yaml.safe_dump(lockspec, f)

    # check if repo is moved to specified commit
    kas.kas('dump test.yml'.split())
    lockspec = yaml.safe_load(capsys.readouterr().out)
    assert lockspec['overrides']['repos']['externalrepo']['commit'] \
        == test_commit

    # update lockfile, check if repo is pinned to other commit
    capsys.readouterr()
    kas.kas('dump --lock --inplace --update test.yml'.split())
    assert test_shortmsg in capsys.readouterr().out
    with open('test.lock.yml', 'r') as f:
        lockspec = yaml.safe_load(f)
        assert lockspec['overrides']['repos']['externalrepo']['commit'] \
            != test_commit

    # check if lockfile gets updated when repo is pinned to a nonexistent
    # commit
    nonexistent_commit = '0000000000000000000000000000000000000000'
    lockspec['overrides']['repos']['externalrepo']['commit'] = \
        nonexistent_commit
    with open('test.lock.yml', 'w') as f:
        yaml.safe_dump(lockspec, f)
    kas.kas('dump --lock --inplace --update test.yml'.split())
    with open('test.lock.yml', 'r') as f:
        lockspec = yaml.safe_load(f)
        assert lockspec['overrides']['repos']['externalrepo']['commit'] \
            != nonexistent_commit


@pytest.mark.dirsfromenv
def test_root_resolve_novcs(monkeykas, tmpdir, capsys):
    tdir = str(tmpdir.mkdir('test_root_resolve_novcs'))
    shutil.rmtree(tdir, ignore_errors=True)
    shutil.copytree('tests/test_commands', tdir)
    monkeykas.chdir(tdir)

    capsys.readouterr()
    kas.kas('dump --resolve-local test-local.yml'.split())
    console = capsys.readouterr()
    assert 'not under version control' in console.err
    data = yaml.safe_load(console.out)
    assert data['repos']['local'] is None


@pytest.mark.dirsfromenv
def test_root_resolve_git(monkeykas, tmpdir, capsys):
    tdir = str(tmpdir.mkdir('test_root_resolve_git'))
    shutil.rmtree(tdir, ignore_errors=True)
    shutil.copytree('tests/test_commands', tdir)
    monkeykas.chdir(tdir)

    upstream_url = 'http://github.com/siemens/kas.git'
    subprocess.check_call(['git', 'init'])
    subprocess.check_call(['git', 'branch', '-m', 'main'])
    subprocess.check_call(['git', 'add', 'oe-init-build-env', '*.yml'])
    subprocess.check_call(['git', 'commit', '-m', 'test'])
    subprocess.check_call(['git', 'remote', 'add', 'origin', upstream_url])
    commit = subprocess.check_output(['git', 'rev-parse', '--verify', 'HEAD'])

    capsys.readouterr()
    kas.kas('dump --resolve-local test-local.yml'.split())
    console = capsys.readouterr()
    data = yaml.safe_load(console.out)
    assert data['repos']['local']['commit'] == commit.decode('utf-8').strip()
    assert data['repos']['local']['url'] == upstream_url

    # make repository dirty
    with open(f'{tdir}/new-file.txt', 'w') as f:
        f.write('test')
    kas.kas('dump --resolve-local test-local.yml'.split())
    console = capsys.readouterr()
    data = yaml.safe_load(console.out)
    assert data['repos']['local']['commit'] == commit.decode('utf-8').strip()
    assert data['repos']['local']['url'] == upstream_url


@pytest.mark.dirsfromenv
def test_root_resolve_hg(monkeykas, tmpdir, capsys):
    tdir = str(tmpdir.mkdir('test_root_resolve_hg'))
    shutil.rmtree(tdir, ignore_errors=True)
    shutil.copytree('tests/test_commands', tdir)
    monkeykas.chdir(tdir)

    upstream_url = 'http://github.com/siemens/kas'
    subprocess.check_call(['hg', 'init'])
    subprocess.check_call(['hg', 'add', 'test-local-hg.yml',
                           'oe-init-build-env'])
    subprocess.check_call(['hg', 'commit', '-m', 'test'])
    with open(f'{tdir}/.hg/hgrc', mode='w') as f:
        f.write(f'[paths]\ndefault = {upstream_url}')
    commit = subprocess.check_output(['hg', 'log', '-r', '.',
                                      '--template', '{node}\n'])

    capsys.readouterr()
    kas.kas('dump --resolve-local test-local-hg.yml'.split())
    console = capsys.readouterr()
    data = yaml.safe_load(console.out)
    assert data['repos']['local']['commit'] == commit.decode('utf-8').strip()
    assert data['repos']['local']['url'] == upstream_url
    assert data['repos']['local']['type'] == 'hg'

    # add all files to repository
    subprocess.check_call(['hg', 'add'])
    subprocess.check_call(['hg', 'commit', '-m', 'test2'])
    kas.kas('dump --resolve-local test-local-hg.yml'.split())


def test_ff_merges(monkeykas, tmpdir):
    """
    This tests check if kas correcly handles fast-forward merges.
    """
    tdir = tmpdir / 'test_commands'
    rdirbare = tdir / 'upstream.git'
    rdirwork = tdir / 'upstream.clone'
    wdirkas = tdir / 'kas'
    for d in [tdir, rdirbare, rdirwork, wdirkas]:
        d.mkdir()

    with open('tests/test_commands/test-ff-merges.yml', 'r') as f:
        kas_input = yaml.safe_load(f)

    shutil.copy('tests/test_commands/oe-init-build-env', rdirwork)
    # create a bare upstream repository
    monkeykas.chdir(rdirbare)
    subprocess.check_call(['git', 'init', '--bare'])
    subprocess.check_call(['git', 'symbolic-ref', 'HEAD', 'refs/heads/main'])
    # create a repository with a branch (working copy)
    monkeykas.chdir(rdirwork)
    subprocess.check_call(['git', 'init'])
    subprocess.check_call(['git', 'branch', '-m', 'main'])
    subprocess.check_call(['git', 'add', 'oe-init-build-env'])
    subprocess.check_call(['git', 'commit', '-m', 'test'])
    # get head of main branch (before merge)
    c_main = subprocess.check_output(['git', 'rev-parse', '--verify', 'HEAD'])
    subprocess.check_call(['git', 'checkout', '-b', 'feature'])
    subprocess.check_call(['touch', 'foo'])
    subprocess.check_call(['git', 'add', 'foo'])
    subprocess.check_call(['git', 'commit', '-m', 'test-feature'])
    # get head of feature branch
    c_feat = subprocess.check_output(['git', 'rev-parse', '--verify', 'HEAD'])
    subprocess.check_call(['git', 'remote', 'add', 'origin', rdirbare])
    subprocess.check_call(['git', 'push', 'origin', 'main', 'feature'])

    # perform initial kas checkout
    monkeykas.chdir(wdirkas)
    kas_input['repos']['upstream']['commit'] = c_main.decode('utf-8').strip()
    kas_input['repos']['upstream']['url'] = str(rdirbare)
    with open('kas.yml', 'w') as f:
        yaml.dump(kas_input, f)
    kas.kas(['checkout', 'kas.yml'])

    # ff merge feature into main
    monkeykas.chdir(rdirwork)
    # checkout main branch with kas
    subprocess.check_call(['git', 'checkout', 'main'])
    subprocess.check_call(['git', 'merge', '--ff-only', 'feature'])
    subprocess.check_call(['git', 'push', 'origin', 'main'])

    # bump commit on main branch in kas project, perform kas checkout
    monkeykas.chdir(wdirkas)
    # checkout main branch with kas again
    kas_input['repos']['upstream']['commit'] = c_feat.decode('utf-8').strip()
    with open('kas.yml', 'w') as f:
        yaml.dump(kas_input, f)
    kas.kas(['checkout', 'kas.yml'])


def test_cmd_not_found(monkeykas, tmpdir):
    cmd = ['/usr/bin/kas-not-exists']
    ret, _ = run_cmd(cmd, tmpdir, os.environ, fail=False)
    assert ret != 0
    with pytest.raises(FileNotFoundError):
        run_cmd(cmd, tmpdir, os.environ, fail=True)



# kas - setup tool for bitbake based projects
#
# Copyright (c) Peter Hatina, 2021
# Copyright (c) Siemens AG, 2021-2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import shutil
import pathlib
import subprocess
import re
import pytest
import json
import logging
from kas import kas
from kas.context import create_global_context
from kas.kasusererror import ArgsCombinationError, EnvSetButNotFoundError
from kas.libcmds import SetupHome
from kas import __version__


@pytest.mark.dirsfromenv
def test_build_dir_is_placed_inside_work_dir_by_default(monkeykas, tmpdir):
    conf_dir = str(tmpdir / 'test_env_variables')
    shutil.copytree('tests/test_environment_variables', conf_dir)

    monkeykas.chdir(conf_dir)

    kas.kas(['checkout', 'test.yml'])

    assert (monkeykas.get_kbd() / 'conf').exists()


def test_build_dir_can_be_specified_by_environment_variable(monkeykas, tmpdir):
    conf_dir = str(tmpdir / 'test_env_variables')
    build_dir = str(tmpdir / 'test_build_dir')
    shutil.copytree('tests/test_environment_variables', conf_dir)
    monkeykas.chdir(conf_dir)

    monkeykas.setenv('KAS_BUILD_DIR', build_dir)
    kas.kas(['checkout', 'test.yml'])

    assert os.path.exists(os.path.join(build_dir, 'conf'))


@pytest.mark.dirsfromenv
def test_ssh_agent_setup(monkeykas, tmpdir, capsys):
    conf_dir = str(tmpdir / 'test_ssh_agent_setup')
    shutil.copytree('tests/test_environment_variables', conf_dir)
    monkeykas.chdir(conf_dir)

    SSH_AUTH_SOCK = '/tmp/ssh-KLTafE/agent.64708'

    with monkeykas.context() as mp:
        envfile = tmpdir / 'env'
        mp.setenv('SSH_AUTH_SOCK', SSH_AUTH_SOCK)
        kas.kas(['shell', '-c', f'env > {envfile}', 'test.yml'])
        env = _get_env_from_file(envfile)
        assert env['SSH_AUTH_SOCK'] == SSH_AUTH_SOCK

    with monkeykas.context() as mp:
        mp.setenv('SSH_AUTH_SOCK', SSH_AUTH_SOCK)
        mp.setenv('SSH_PRIVATE_KEY', 'id_rsa')
        with pytest.raises(ArgsCombinationError):
            kas.kas(['checkout', 'test.yml'])

    privkey_file = f'{tmpdir}/id_ecdsa_test'
    genkey_cmd = ['ssh-keygen', '-f', privkey_file, '-N', '', '-t', 'ecdsa']
    subprocess.check_call(genkey_cmd)
    # ensure we also get the info messages
    log = kas.logging.getLogger()
    log.setLevel(kas.logging.INFO)
    # flush the captured output
    capsys.readouterr()
    with monkeykas.context() as mp:
        mp.setenv('SSH_PRIVATE_KEY_FILE', privkey_file)
        kas.kas(['checkout', 'test.yml'])
        out = capsys.readouterr().err
        assert 'adding SSH key from file' in out
        assert 'ERROR' not in out

    with monkeykas.context() as mp:
        privkey = pathlib.Path(privkey_file).read_text()
        mp.setenv('SSH_PRIVATE_KEY', privkey)
        kas.kas(['checkout', 'test.yml'])
        out = capsys.readouterr().err
        assert 'adding SSH key from env-var' in out
        assert 'ERROR' not in out


def _get_env_from_file(filename):
    env = {}
    with filename.open() as f:
        for line in f.readlines():
            key, val = line.split("=", 1)
            env[key] = val.strip()
    return env


def _test_env_section_export(monkeykas, tmpdir, bb_env_var, bb_repo):
    conf_dir = tmpdir / 'test_env_variables'
    env_out = conf_dir / 'env_out'
    bb_env_out = conf_dir / 'bb_env_out'
    init_build_env = conf_dir / 'oe-init-build-env'

    shutil.copytree('tests/test_environment_variables', conf_dir)
    monkeykas.chdir(conf_dir)
    monkeykas.setenv('KAS_CLONE_DEPTH', '1')
    monkeykas.setenv('KAS_PREMIRRORS', 'https://git\\.openembedded\\.org/ '
                     'https://github.com/openembedded/\n')
    # resolve after we chdir'd
    kas_wd = monkeykas.get_kwd()
    kas_bd = monkeykas.get_kbd()

    # Overwrite oe-init-build-env script
    # BB_ENV_* filter variables are only exported by
    # kas when they are already exported in the setup environment script
    script = """#!/bin/sh
    export %s="FOO"
    export PATH="%s/%s/bin:${PATH}"
    """ % (bb_env_var, str(kas_wd), bb_repo)
    init_build_env.write_text(script, encoding=None)
    init_build_env.chmod(0o775)

    # Before executing bitbake, first get the bitbake.conf
    kas.kas(['checkout', 'test_env.yml'])
    shutil.copy(kas_wd / bb_repo / 'conf' / 'bitbake.conf',
                kas_bd / 'conf' / 'bitbake.conf')

    kas.kas(['shell', '-c', 'env > %s' % env_out, 'test_env.yml'])
    kas.kas(['shell', '-c', 'bitbake -e > %s' % bb_env_out, 'test_env.yml'])

    # Check kas environment
    test_env = _get_env_from_file(env_out)

    # Variables with 'None' assigned should not be added to environment
    try:
        _ = test_env['TESTVAR_WHITELIST']
        assert False
    except KeyError:
        assert True

    assert test_env['TESTVAR_DEFAULT_VAL'] == 'BAR'
    assert 'TESTVAR_WHITELIST' in test_env[bb_env_var]

    # Check bitbake's environment
    test_bb_env = {}
    with bb_env_out.open() as f:
        for line in f.readlines():
            if re.match(r'^#', line):
                continue
            match = re.match(r'(^[a-zA-Z0-9_]+)=\"([a-zA-Z0-9_ ]+)\"', line)
            if match:
                key, val = match.group(1), match.group(2)
                test_bb_env[key] = val.strip()

    assert 'TESTVAR_WHITELIST' in test_bb_env[bb_env_var]
    assert test_bb_env["TESTVAR_DEFAULT_VAL"] == "BAR"


# BB_ENV_EXTRAWHITE is deprecated but may still be used
@pytest.mark.online
def test_env_section_export_bb_extra_white(monkeykas, tmpdir):
    _test_env_section_export(monkeykas, tmpdir, 'BB_ENV_EXTRAWHITE',
                             'bitbake_old')


@pytest.mark.online
def test_env_section_export_bb_env_passthrough_additions(monkeykas, tmpdir):
    _test_env_section_export(monkeykas, tmpdir, 'BB_ENV_PASSTHROUGH_ADDITIONS',
                             'bitbake_new')


def test_managed_env_detection(monkeykas):
    with monkeykas.context() as mp:
        mp.setenv('GITLAB_CI', 'true')
        ctx = create_global_context([])
        me = ctx.managed_env
        assert bool(me)
        assert str(me) == 'GitLab CI'
    with monkeykas.context() as mp:
        mp.setenv('GITHUB_ACTIONS', 'true')
        ctx = create_global_context([])
        me = ctx.managed_env
        assert bool(me)
        assert str(me) == 'GitHub Actions'
    with monkeykas.context() as mp:
        mp.setenv('REMOTE_CONTAINERS', 'true')
        mp.setenv('REMOTE_CONTAINERS_FOO', 'bar')
        ctx = create_global_context([])
        me = ctx.managed_env
        assert bool(me)
        assert str(me) == 'VSCode Remote Containers'
        assert ctx.environ['REMOTE_CONTAINERS_FOO'] == 'bar'


@pytest.mark.dirsfromenv
def test_env_file_processing(monkeykas, tmpdir):
    ctx = create_global_context([])
    rcfiles = [
        ('NETRC_FILE', '.netrc'),
        ('NPMRC_FILE', '.npmrc'),
        ('REGISTRY_AUTH_FILE', '.docker/config.json'),
        ('AWS_CONFIG_FILE', '.aws/config'),
        ('AWS_WEB_IDENTITY_TOKEN_FILE', '.aws/web_identity_token')
        # no need to test gitconfig, as tested in test_gitconfig
    ]
    for name, file in rcfiles:
        with monkeykas.context() as mp:
            local_file = tmpdir / '.rcfile'
            with open(local_file, 'w') as f:
                if name in ['NETRC_FILE', 'NPMRC_FILE']:
                    f.write('machine foo\nlogin bar\npassword baz\n')
                elif name in ['AWS_CONFIG_FILE',
                              'AWS_WEB_IDENTITY_TOKEN_FILE']:
                    f.write('[profile foo]\nregion=bar\n')
                elif name == 'REGISTRY_AUTH_FILE':
                    json.dump({'auths': {'example.com': {'auth': 'foo'}}}, f)
                else:
                    f.write('foo: bar\n')
            mp.setenv(name, str(local_file))
            if name in ['AWS_CONFIG_FILE', 'AWS_WEB_IDENTITY_TOKEN_FILE']:
                mp.setenv('AWS_SHARED_CREDENTIALS_FILE', str(local_file))
                mp.setenv('AWS_ROLE_ARN', 'arn:foo')
                if name == 'AWS_WEB_IDENTITY_TOKEN_FILE':
                    mp.setenv('AWS_CONFIG_FILE', str(local_file))
            elif name == 'REGISTRY_AUTH_FILE':
                mp.setenv('CI_JOB_TOKEN', 'foo')
                mp.setenv('CI_REGISTRY_USER', 'reg-user')
                mp.setenv('CI_REGISTRY', 'example.com')
            sh = SetupHome()
            sh.execute(ctx)
            assert (pathlib.Path(sh.tmpdirname) / pathlib.Path(file)).exists()


def test_env_set_but_not_existing(monkeykas):
    with monkeykas.context() as mp:
        mp.setenv('NETRC_FILE', '/path/does/not/exist')
        ctx = create_global_context([])
        with pytest.raises(EnvSetButNotFoundError):
            SetupHome().execute(ctx)


def test_kas_container_version(monkeykas, caplog):
    caplog.set_level(logging.WARNING)
    with monkeykas.context() as mp:
        # not a kas-container call
        create_global_context([])
        assert 'versions do not match' not in caplog.text
        caplog.clear()
    with monkeykas.context() as mp:
        # kas-container call from matching script
        mp.setenv('KAS_CONTAINER_SCRIPT_VERSION', __version__)
        create_global_context([])
        assert 'versions do not match' not in caplog.text
        caplog.clear()
    with monkeykas.context() as mp:
        # kas-container call from older/newer script
        mp.setenv('KAS_CONTAINER_SCRIPT_VERSION', '0.0')
        create_global_context([])
        assert 'versions do not match' in caplog.text
        caplog.clear()



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2017-2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This module contains the core implementation of kas.
"""

import argparse
import re
import os
import sys
import logging
import tempfile
import asyncio
import errno
import hashlib
import hmac
import pathlib
import platform
import shutil
import signal
import stat
from subprocess import Popen, PIPE, run as subprocess_run
from urllib.parse import quote
from .context import get_context
from .kasusererror import KasUserError, CommandExecError
from .configschema import CONFIGSCHEMA

__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens AG, 2017-2018'


class InitBuildEnvError(KasUserError):
    """
    Error related to the OE / ISAR environment setup scripts
    """
    pass


class EnvNotValidError(KasUserError):
    """
    The caller environment is not suited for the requested operation
    """
    pass


class TaskExecError(KasUserError):
    """
    Similar to :class:`kas.kasusererror.CommandExecError` but for kas
    internal tasks
    """
    def __init__(self, command, ret_code):
        self.ret_code = ret_code
        super().__init__(f'{command} failed: error code {ret_code}')


class LogOutput:
    """
        Handles the log output of executed applications
    """
    def __init__(self, live):
        self.live = live
        self.stdout = []
        self.stderr = []

    def log_stdout(self, line):
        """
            This method is called when a line is received over stdout.
        """
        if self.live:
            logging.info(line.strip())
        self.stdout.append(line)

    def log_stderr(self, line):
        """
            This method is called when a line is received over stderr.
        """
        if self.live:
            logging.error(line.strip())
        self.stderr.append(line)


async def _read_stream(stream, callback):
    """
        This asynchronous method reads from the output stream of the
        application and transfers each line to the callback function.
    """
    while True:
        line = await stream.readline()
        try:
            line = line.decode('utf-8')
        except UnicodeDecodeError as err:
            logging.warning('Could not decode line from stream, ignoring: %s',
                            err)
        if line:
            callback(line)
        else:
            break


def _filter_stderr(capture_stderr, ret, out, err=None):
    if capture_stderr:
        return (ret, out, err or '')
    else:
        return (ret, out)


def _report_cmd_error(ret, cwd, cmdstr, cmd, stderr):
    if stderr:
        msg = f'Command "{cwd}$ {cmdstr}" failed:\n{stderr}'
        logging.error(msg.rstrip('\n'))
    raise CommandExecError(cmd, ret)


async def run_cmd_async(cmd, cwd, env=None, fail=True, liveupdate=False,
                        capture_stderr=False):
    """
        Run a command asynchronously.
    """

    env = env or get_context().environ
    cmdstr = ' '.join(cmd)
    logging.debug('%s$ %s', cwd, cmdstr)

    logo = LogOutput(liveupdate)

    orig_fd = signal.set_wakeup_fd(-1, warn_on_full_buffer=False)
    signal.set_wakeup_fd(orig_fd, warn_on_full_buffer=False)

    try:
        process = await asyncio.create_subprocess_exec(
            *cmd,
            cwd=cwd,
            env=env,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            preexec_fn=os.setpgrp)
    except FileNotFoundError as ex:
        if fail:
            raise ex
        return _filter_stderr(capture_stderr, errno.ENOENT, str(ex))
    except PermissionError as ex:
        if fail:
            raise ex
        return _filter_stderr(capture_stderr, errno.EPERM, str(ex))

    # Process termination is a complicated thing. We need to ensure that
    # the event-loop ThreadedChildWatcher thread fires before the loop is
    # terminated. The best we can do it to ask the process to terminate
    # (SIGINT) and wait for it. We need to shield the process wait to avoid
    # that it is killed by the cancellation of the task, as we want a
    # controlled termination. Forced terminations can leak an orphaned process.
    # https://github.com/pytest-dev/pytest-asyncio/issues/708#issuecomment-1868488942
    tasks = [
        asyncio.ensure_future(_read_stream(process.stdout, logo.log_stdout)),
        asyncio.ensure_future(_read_stream(process.stderr, logo.log_stderr))
    ]

    try:
        await asyncio.gather(*[asyncio.shield(t) for t in tasks])
        ret = await asyncio.shield(process.wait())
    except asyncio.CancelledError:
        try:
            process.terminate()
        except ProcessLookupError:
            # Process already exited between the cancel and us reaching here.
            pass
        logging.debug('Command "%s" cancelled', cmdstr)
        await process.wait()
        raise

    if ret and fail:
        _report_cmd_error(ret, cwd, cmdstr, cmd, logo.stderr)

    return _filter_stderr(capture_stderr, ret,
                          ''.join(logo.stdout), ''.join(logo.stderr))


def run_cmd(cmd, cwd, env=None, fail=True, capture_stderr=False):
    """
        Runs a command synchronously.
    """
    env = env or get_context().environ
    cmdstr = ' '.join(cmd)
    logging.debug('%s$ %s', cwd, cmdstr)

    try:
        ret = subprocess_run(cmd, env=env, cwd=cwd, stdout=PIPE, stderr=PIPE)
        if ret.returncode and fail:
            _report_cmd_error(ret.returncode, cwd, cmdstr, cmd,
                              ret.stderr.decode('utf-8'))
    except FileNotFoundError as ex:
        if fail:
            raise ex
        return _filter_stderr(capture_stderr, errno.ENOENT, str(ex))
    except PermissionError as ex:
        if fail:
            raise ex
        return _filter_stderr(capture_stderr, errno.EPERM, str(ex))
    return _filter_stderr(capture_stderr, ret.returncode,
                          ret.stdout.decode('utf-8'),
                          ret.stderr.decode('utf-8'))


def find_program(paths, name):
    """
        Find a file within the paths array and returns its path.
    """
    for path in paths.split(os.pathsep):
        prg = os.path.join(path, name)
        if os.path.isfile(prg):
            return prg
    return None


def repos_fetch(repos):
    """
        Fetches the list of repositories to the kas_work_dir.

        .. note:: termination point of the asyncio event loop.
    """
    if len(repos) == 0:
        return

    tasks = []
    for repo in repos:
        tasks.append(asyncio.ensure_future(repo.fetch_async()))

    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(asyncio.gather(*tasks))
    except CommandExecError as e:
        [t.cancel() for t in tasks]
        raise TaskExecError('fetch repos', e.ret_code)
    except KasUserError:
        [t.cancel() for t in tasks]
        raise


def repos_apply_patches(repos):
    """
        Applies the patches to the repositories.

        .. note:: termination point of the asyncio event loop.
    """
    if len(repos) == 0:
        return

    tasks = []
    for repo in repos:
        tasks.append(asyncio.ensure_future(repo.apply_patches_async()))

    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(asyncio.gather(*tasks))
    except CommandExecError as e:
        [t.cancel() for t in tasks]
        raise TaskExecError('apply patches', e.ret_code)
    except KasUserError:
        [t.cancel() for t in tasks]
        raise


def get_buildtools_dir():
    # Set the dest. directory for buildtools's setup
    env_path = os.environ.get("KAS_BUILDTOOLS_DIR")
    if env_path:
        return pathlib.Path(env_path).resolve()

    # defaults to KAS_BUILD_DIR/buildtools
    return (pathlib.Path(get_context().build_dir) / 'buildtools').resolve()


def get_buildtools_filename():
    arch = platform.machine()
    ctx = get_context()

    conf_buildtools = ctx.config.get_buildtools()
    version = conf_buildtools['version']
    if 'filename' in conf_buildtools:
        filename = conf_buildtools['filename']
    else:
        filename = (
            f"{arch}-buildtools-extended-"
            f"nativesdk-standalone-{version}.sh"
        )

    return filename


def get_buildtools_path():
    return get_buildtools_dir() / get_buildtools_filename()


def get_buildtools_url():
    ctx = get_context()
    conf_buildtools = ctx.config.get_buildtools()
    filename = get_buildtools_filename()
    version = conf_buildtools['version']

    if 'base_url' in conf_buildtools:
        base_url = conf_buildtools['base_url']
    else:
        default = (
            CONFIGSCHEMA['properties']['buildtools']['properties']
            ['base_url']['default']
        )
        base_url = f"{default}/yocto-{version}/buildtools/"

    return f"{base_url}/{quote(filename)}"


def check_sha256sum(filename, expected_checksum):
    hash_sha256 = hashlib.sha256()
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(8192), b""):
            hash_sha256.update(chunk)

    actual_checksum = hash_sha256.hexdigest()
    logging.info(
        f"Buildtools installer's checksum (sha256) is: "
        f"{actual_checksum}"
    )

    return hmac.compare_digest(actual_checksum, expected_checksum)


def download_buildtools():
    ctx = get_context()
    conf_buildtools = ctx.config.get_buildtools()
    version = conf_buildtools['version']
    buildtools_dir = get_buildtools_dir()

    # Enable extended buildtools tarball
    buildtools_url = get_buildtools_url()
    tmpbuildtools = get_buildtools_path()

    logging.info(f"Downloading Buildtools {version}")
    # Download installer
    fetch_cmd = ['wget', '-q', '-O', str(tmpbuildtools), buildtools_url]
    (ret, _) = run_cmd(fetch_cmd, cwd=ctx.kas_work_dir)
    if ret != 0:
        raise InitBuildEnvError("Could not download buildtools installer")

    # Check if the installer's sha256sum matches
    if not check_sha256sum(tmpbuildtools, conf_buildtools['sha256sum']):
        raise InitBuildEnvError(
            "sha256sum mismatch: installer may be corrupted"
        )

    # Make installer executable
    st = tmpbuildtools.stat()
    tmpbuildtools.chmod(st.st_mode | stat.S_IEXEC)

    # Run installer (in an isolated environment)
    installer_cmd = [str(tmpbuildtools), '-d', str(buildtools_dir), '-y']
    env = {'PATH': '/usr/sbin:/usr/bin:/sbin:/bin'}
    (ret, _) = run_cmd(installer_cmd, cwd=ctx.kas_work_dir, env=env)
    if ret != 0:
        raise InitBuildEnvError("Could not run buildtools installer")


def get_buildtools_version():
    try:
        version_file = list(get_buildtools_dir().glob("version-*"))
        if len(version_file) != 1:
            raise ValueError("Invalid number of version files")

        with version_file[0].resolve().open('r') as f:
            lines = f.readlines()
            for line in lines:
                if line.startswith("Distro Version"):
                    return lines[1].split(':', 1)[1].strip()
    except Exception as e:
        logging.warning(f"Unable to read buildtools version: {e}")

    return -1


def get_build_environ(build_system):
    """
        Creates the build environment variables.
    """
    # nasty side effect function: running oe/isar-init-build-env also
    # creates the conf directory

    init_repo = None
    if build_system in ['openembedded', 'oe']:
        scripts = ['oe-init-build-env']
    elif build_system == 'isar':
        scripts = ['isar-init-build-env']
    else:
        scripts = ['oe-init-build-env', 'isar-init-build-env']
    permutations = \
        [(repo, script) for repo in get_context().config.get_repos()
         for script in scripts]
    for (repo, script) in permutations:
        if os.path.exists(repo.path + '/' + script):
            if init_repo:
                raise InitBuildEnvError(
                    'Multiple init scripts found '
                    f'({repo.name} vs. {init_repo.name}). '
                    'Resolve ambiguity by removing one of the repos')

            init_repo = repo
            init_script = script
    if not init_repo:
        raise InitBuildEnvError('Did not find any init-build-env script')

    conf_buildtools = get_context().config.get_buildtools()
    buildtools_env = ""

    if conf_buildtools:
        # Create the dest. directory if it doesn't exist
        buildtools_dir = get_buildtools_dir()
        buildtools_dir.mkdir(parents=True, exist_ok=True)

        if not any(buildtools_dir.iterdir()):
            # Directory is empty, try to fetch from upstream
            logging.info(f"Buildtools ({buildtools_dir}): directory is empty")
            download_buildtools()
        else:
            # Fetch buildtools when versions differ in non-empty dir
            found_version = get_buildtools_version()
            if found_version != conf_buildtools['version']:
                logging.warning("Buildtools: version mismatch")
                logging.info(f"Required version: {conf_buildtools['version']}")
                logging.info(f"Found version: {found_version}")
                shutil.rmtree(os.path.realpath(buildtools_dir))
                os.makedirs(os.path.realpath(buildtools_dir))
                download_buildtools()

        envfiles = list(get_buildtools_dir().glob("environment-setup-*"))
        if len(envfiles) == 1:
            # Ignore missing pkg-config error until oe-core fix is merged
            buildtools_env = (
                "source {} || true\n".format(envfiles[0].resolve())
            )
        else:
            logging.error(
                f"Expected 1 environment setup file, found {len(envfiles)}."
                "Invalid or misconfigured buildtools package."
            )
            return -1

    with tempfile.TemporaryDirectory() as temp_dir:
        if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
            init_script_log = pathlib.Path(temp_dir) / '.init_script.log'
        else:
            init_script_log = '/dev/null'
        script = f"""#!/bin/bash
        set -e
        {buildtools_env}
        source {init_script} $1 > {init_script_log}
        env
        """

        get_bb_env_file = pathlib.Path(temp_dir) / "get_bb_env"
        get_bb_env_file.write_text(script)
        get_bb_env_file.chmod(0o775)

        env = {}
        env['PATH'] = '/usr/sbin:/usr/bin:/sbin:/bin'

        (_, output) = run_cmd([str(get_bb_env_file), get_context().build_dir],
                              cwd=init_repo.path, env=env)
        if init_script_log != '/dev/null':
            with open(init_script_log) as log:
                msg = f'{init_script} output:\n'
                for line in log.readlines():
                    msg += line
                logging.debug(msg.rstrip('\n'))

    env = {}
    for line in output.splitlines():
        try:
            (key, val) = line.split('=', 1)
            env[key] = val
        except ValueError:
            pass

    conf_env = get_context().config.get_environment()

    env_vars = ['SSTATE_DIR', 'SSTATE_MIRRORS', 'DL_DIR', 'TMPDIR']
    env_vars.extend(conf_env)

    env.update(conf_env)

    if 'BB_ENV_PASSTHROUGH_ADDITIONS' in env:
        passthrough_additions = env['BB_ENV_PASSTHROUGH_ADDITIONS'] + ' ' + \
            ' '.join(env_vars)
        env.update({'BB_ENV_PASSTHROUGH_ADDITIONS': passthrough_additions})
    elif 'BB_ENV_EXTRAWHITE' in env:
        extra_white = env['BB_ENV_EXTRAWHITE'] + ' ' + ' '.join(env_vars)
        env.update({'BB_ENV_EXTRAWHITE': extra_white})

    env_vars.extend(['SSH_AUTH_SOCK',
                     'SHELL', 'TERM',
                     'GIT_PROXY_COMMAND', 'NO_PROXY'])

    for env_var in env_vars:
        if env_var in os.environ:
            env[env_var] = os.environ[env_var]

    # filter out 'None' values
    env = {k: v for (k, v) in env.items() if v is not None}

    return env


def ssh_add_key_file(env, key_path):
    """
        Adds an ssh key file to the ssh-agent
    """
    with open(key_path) as f:
        key = f.read()
        ssh_add_key(env, key)


def ssh_add_key(env, key):
    """
        Adds an ssh key to the ssh-agent
    """
    # The ssh-agent needs the key to end with a newline, otherwise it
    # unhelpfully prompts for a password
    if not key.endswith('\n'):
        key += '\n'

    process = Popen(['ssh-add', '-'], stdin=PIPE, stdout=None,
                    stderr=PIPE, env=env)
    (_, error) = process.communicate(input=str.encode(key))
    if process.returncode and error:
        logging.error('failed to add ssh key: %s', error.decode('utf-8'))


def ssh_cleanup_agent():
    """
        Removes the identities and stops the ssh-agent instance
    """
    ctx = get_context()
    # remove the identities
    (ret, _) = run_cmd(['ssh-add', '-D'], cwd=ctx.kas_work_dir,
                       env=ctx.environ, fail=False)
    if ret != 0:
        logging.error('failed to delete SSH identities')

    # stop the ssh-agent
    (ret, _) = run_cmd(['ssh-agent', '-k'], cwd=ctx.kas_work_dir,
                       env=ctx.environ, fail=False)
    if ret != 0:
        logging.error('failed to stop SSH agent')


def ssh_setup_agent(envkeys=None):
    """
        Starts the ssh-agent
    """
    ctx = get_context()
    env = ctx.environ
    envkeys = envkeys or ['SSH_PRIVATE_KEY', 'SSH_PRIVATE_KEY_FILE']
    (_, output) = run_cmd(['ssh-agent', '-s'], env=env,
                          cwd=ctx.kas_work_dir)
    for line in output.split('\n'):
        matches = re.search(r"(\S+)\=(\S+)\;", line)
        if matches:
            env[matches.group(1)] = matches.group(2)

    found = False
    for envkey in envkeys:
        if envkey == 'SSH_PRIVATE_KEY_FILE':
            key_path = os.environ.get(envkey)
            if key_path:
                found = True
                logging.info(f"adding SSH key from file '{key_path}'")
                ssh_add_key_file(env, key_path)
        else:
            key = os.environ.get(envkey)
            if key:
                found = True
                logging.info("adding SSH key from env-var 'SSH_PRIVATE_KEY'")
                ssh_add_key(env, key)

    if found is not True:
        warning = "None of the following environment keys were set: " + \
            ", ".join(envkeys)
        logging.warning(warning)


def ssh_no_host_key_check():
    """
        Disables ssh host key check
    """
    home = os.path.expanduser('~')
    ssh_dir = home + '/.ssh'
    if not os.path.exists(ssh_dir):
        os.mkdir(ssh_dir)
    ssh_config = ssh_dir + "/config"
    generated_content = 'Host *\n\tStrictHostKeyChecking no\n\n'
    try:
        with open(ssh_config, 'x') as fds:
            fds.write(generated_content)
    except FileExistsError:
        with open(ssh_config, 'r') as fds:
            content = fds.read()
        if content != generated_content:
            logging.warning("%s already exists, "
                            "not touching it to disable StrictHostKeyChecking",
                            ssh_config)


def setup_parser_common_args(parser):
    from kas.libcmds import Macro

    setup_cmds = [str(s) for (s, _) in Macro().setup_commands]
    parser.add_argument('--skip',
                        help='Skip build steps. To skip more than one step, '
                        'use this argument multiple times.',
                        default=[],
                        action='append',
                        metavar='STEP',
                        choices=setup_cmds)
    parser.add_argument('--force-checkout', action='store_true',
                        help='Always checkout the desired commit/branch/tag '
                        'of each repository, discarding any local changes')
    parser.add_argument('--update', action='store_true',
                        help='Pull new upstream changes to the desired '
                        'branch even if it is already checked out locally')


def setup_parser_config_arg(parser):
    parser.add_argument('config',
                        help='Config file(s), separated by colon. Using '
                        '.config.yaml in KAS_WORK_DIR if none is '
                        'specified.',
                        nargs='?')


def setup_parser_preserve_env_arg(parser):
    parser.add_argument('-E', '--preserve-env',
                        help='Keep current user environment block',
                        action='store_true')


class ExtendConstAction(argparse._AppendConstAction):
    """Add an 'extend_const' action similar to 'append_const'.

    Based on the existing 'append_const' and 'extend' actions.
    """
    def __init__(self, option_strings, dest, const, default=None,
                 required=False, help=None, metavar=None):
        super(argparse._AppendConstAction, self).__init__(
            option_strings=option_strings, dest=dest, nargs=0, const=const,
            default=default, required=required, help=help, metavar=metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        items = getattr(namespace, self.dest, None)
        if items is None:
            items = []

        if isinstance(items, list):
            items = items[:]
        else:
            import copy
            items = copy.copy(items)

        items.extend(self.const)
        setattr(namespace, self.dest, items)


def setup_parser_keep_config_unchanged_arg(parser):
    # Skip the tasks which would change the config of the build
    # environment
    steps = [
        'setup_dir',
        'finish_setup_repos',
        'repos_checkout',
        'repos_apply_patches',
        'write_bbconfig',
    ]
    parser.add_argument('-k', '--keep-config-unchanged',
                        help='Skip steps that change the configuration',
                        action=ExtendConstAction,
                        dest='skip',
                        const=steps)


def run_handle_preserve_env_arg(ctx, os, args, SetupHome):
    if args.preserve_env:
        # Warn if there's any settings that setup_home would apply
        # but are now ignored
        for var in SetupHome.ENV_VARS:
            if var in os.environ:
                logging.warning('Environment variable "%s" ignored '
                                'because user environment is being used',
                                var)

        if not os.isatty(sys.stdout.fileno()):
            raise EnvNotValidError(
                '--preserve-env can only be run from a tty')

        ctx.environ = os.environ.copy()

        logging.warning("Preserving the current environment block may "
                        "have unintended side effects on the build.")



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2019
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest
import shutil
import yaml
import subprocess
import os
from pathlib import Path
from kas import kas
from kas.repos import RepoRefError, Repo


def run_cmd(cmd, cwd=None, fail=True):
    """
        Run a command and return the return code and output.
        Replacement for kas internal run_cmd function which
        cannot be used outside of kas as there is no event loop.
    """
    kas_wd = Path(os.environ.get('KAS_WORK_DIR', '.'))
    _cwd = Path(cwd or '.')
    if not _cwd.is_absolute():
        _cwd = kas_wd / _cwd
    try:
        output = subprocess.check_output(cmd, cwd=_cwd)
        return (0, output.decode('utf-8'))
    except subprocess.CalledProcessError as e:
        if fail:
            raise e
        return (e.returncode, e.output.decode('utf-8'))


@pytest.mark.online
def test_refspec_switch(monkeykas, tmpdir):
    """
        Test that the local git clone is correctly updated when switching
        between a commit hash refspec and a branch refspec.
    """
    tdir = str(tmpdir / 'test_refspec_switch')
    shutil.copytree('tests/test_refspec', tdir)
    monkeykas.chdir(tdir)

    kas.kas(['shell', 'test.yml', '-c', 'true'])
    (rc, output) = run_cmd(['git', 'symbolic-ref', '-q', 'HEAD'], cwd='kas',
                           fail=False)
    assert rc != 0
    assert output.strip() == ''
    (rc, output) = run_cmd(['git', 'rev-parse', 'HEAD'], cwd='kas',
                           fail=False)
    assert rc == 0
    assert output.strip() == '907816a5c4094b59a36aec12226e71c461c05b77'
    (rc, output) = run_cmd(['git', 'symbolic-ref', '-q', 'HEAD'], cwd='kas2',
                           fail=False)
    assert rc == 0
    assert output.strip() == 'refs/heads/master'
    (rc, output) = run_cmd(['git', 'tag', '--points-at', 'HEAD'], cwd='kas3',
                           fail=False)
    assert rc == 0
    assert output.strip() == '3.0.1'

    kas.kas(['shell', 'test2.yml', '-c', 'true'])
    (rc, output) = run_cmd(['git', 'symbolic-ref', '-q', 'HEAD'], cwd='kas',
                           fail=False)
    assert rc == 0
    assert output.strip() == 'refs/heads/master'
    (rc, output) = run_cmd(['git', 'symbolic-ref', '-q', 'HEAD'], cwd='kas2',
                           fail=False)
    assert rc != 0
    assert output.strip() == ''
    (rc, output) = run_cmd(['git', 'rev-parse', 'HEAD'], cwd='kas2',
                           fail=False)
    assert rc == 0
    assert output.strip() == '907816a5c4094b59a36aec12226e71c461c05b77'
    (rc, output) = run_cmd(['git', 'symbolic-ref', '-q', 'HEAD'], cwd='kas3',
                           fail=False)
    assert rc == 0
    assert output.strip() == 'refs/heads/master'
    (rc, output) = run_cmd(['git', 'tag', '--points-at', 'HEAD'], cwd='kas4',
                           fail=False)
    assert rc == 0
    assert output.strip() == '2.6.3'


@pytest.mark.online
def test_refspec_absolute(monkeykas, tmpdir):
    """
        Test that the local git clone works when a absolute refspec
        is given.
    """
    tdir = str(tmpdir / 'test_refspec_absolute')
    shutil.copytree('tests/test_refspec', tdir)
    monkeykas.chdir(tdir)

    kas.kas(['shell', 'test3.yml', '-c', 'true'])
    (rc, output) = run_cmd(['git', 'symbolic-ref', '-q', 'HEAD'],
                           cwd='kas_abs', fail=False)
    assert rc == 0
    assert output.strip() == 'refs/heads/master'
    (rc, output_kas_abs) = run_cmd(['git', 'rev-parse', 'HEAD'],
                                   cwd='kas_abs', fail=False)
    assert rc == 0
    (rc, output_kas_rel) = run_cmd(['git', 'rev-parse', 'HEAD'],
                                   cwd='kas_rel', fail=False)
    assert rc == 0
    assert output_kas_abs.strip() == output_kas_rel.strip()
    (rc, output) = run_cmd(['git', 'tag', '--points-at', 'HEAD'],
                           cwd='kas_tag_abs', fail=False)
    assert rc == 0
    assert output.strip() == '3.0.1'


def test_url_no_refspec(monkeykas, tmpdir):
    """
        Test that a repository with url but no refspec raises an error.
    """
    tdir = str(tmpdir / 'test_url_no_refspec')
    shutil.copytree('tests/test_refspec', tdir)
    monkeykas.chdir(tdir)
    with pytest.raises(RepoRefError):
        kas.kas(['shell', 'test4.yml', '-c', 'true'])


def test_commit_refspec_mix(monkeykas, tmpdir):
    """
        Test that mixing legacy refspec with commit/branch raises errors.
    """
    tdir = str(tmpdir / 'test_commit_refspec_mix')
    shutil.copytree('tests/test_refspec', tdir)
    monkeykas.chdir(tdir)
    with pytest.raises(RepoRefError):
        kas.kas(['shell', 'test5.yml', '-c', 'true'])
    with pytest.raises(RepoRefError):
        kas.kas(['shell', 'test6.yml', '-c', 'true'])
    with pytest.raises(RepoRefError):
        kas.kas(['shell', 'test7.yml', '-c', 'true'])


@pytest.mark.online
def test_tag_commit_do_not_match(monkeykas, tmpdir):
    """
        Test that giving tag and commit that do not match raises an error.
    """
    tdir = str(tmpdir / 'test_tag_commit_do_not_match')
    shutil.copytree('tests/test_refspec', tdir)
    monkeykas.chdir(tdir)
    with pytest.raises(RepoRefError):
        kas.kas(['shell', 'test8.yml', '-c', 'true'])


@pytest.mark.online
def test_unsafe_tag_warning(capsys, monkeykas, tmpdir):
    """
        Test that using tag without commit issues a warning, but only once.
    """
    tdir = str(tmpdir / 'test_unsafe_tag_warning')
    shutil.copytree('tests/test_refspec', tdir)
    monkeykas.chdir(tdir)
    # needs to be reset in case other tests ran before
    Repo.__no_commit_tag_warned__ = []
    kas.kas(['shell', 'test2.yml', '-c', 'true'])
    assert capsys.readouterr().err.count(
        'Using tag without commit for repository "kas4" is unsafe as tags '
        'are mutable.') == 1


@pytest.mark.online
def test_tag_branch_same_name(capsys, monkeykas, tmpdir):
    """
        Test that kas uses the tag if a branch has the same name as the tag.
    """
    tdir = str(tmpdir / 'test_tag_branch_same_name')
    shutil.copytree('tests/test_refspec', tdir)
    monkeykas.chdir(tdir)

    # Checkout the repositories
    kas.kas(['shell', 'test.yml', '-c', 'true'])

    # In kas3: create a branch named "3.0.1" on master HEAD
    # A tag named "3.0.1" already exists on an old commit from 2022
    (rc, output) = run_cmd(['git', 'switch', 'master'], cwd='kas3',
                           fail=False)
    assert rc == 0
    (rc, output) = run_cmd(['git', 'branch', '3.0.1'], cwd='kas3',
                           fail=False)
    assert rc == 0

    # In kas4: create a tag named "master" on existing 2.6.3 tag
    (rc, output) = run_cmd(['git', 'checkout', '2.6.3'], cwd='kas4',
                           fail=False)
    assert rc == 0
    (rc, output) = run_cmd(['git', 'tag', 'master'], cwd='kas4',
                           fail=False)
    assert rc == 0

    # Checkout the repositories again
    kas.kas(['shell', 'test.yml', '-c', 'true'])

    # Check the commit hashes
    (rc, output) = run_cmd(['git', 'rev-parse', 'HEAD'], cwd='kas3',
                           fail=False)
    assert rc == 0
    assert output.strip() == '229310958b17dc2b505b789c1cc1d0e2fddccc44'

    (rc, output) = run_cmd(['git', 'rev-parse', 'HEAD'], cwd='kas4',
                           fail=False)
    assert rc == 0

    (rc, output2) = run_cmd(['git', 'rev-parse', 'refs/heads/master'],
                            cwd='kas4', fail=False)
    assert rc == 0
    assert output.strip() == output2.strip()


@pytest.mark.online
def test_refspec_warning(capsys, monkeykas, tmpdir):
    """
        Test that using legacy refspec issues a warning, but only once.
    """
    tdir = str(tmpdir / 'test_refspec_warning')
    shutil.copytree('tests/test_refspec', tdir)
    monkeykas.chdir(tdir)
    # needs to be reset in case other tests ran before
    Repo.__legacy_refspec_warned__ = []
    kas.kas(['shell', 'test2.yml', '-c', 'true'])
    assert capsys.readouterr().err.count(
        'Using deprecated refspec for repository "kas2".') == 1


@pytest.mark.online
def test_branch_and_tag(monkeykas, tmpdir, mercurial):
    """
        Test if error is raised when branch and tag are set.
    """
    tdir = str(tmpdir / 'test_branch_and_tag')
    shutil.copytree('tests/test_refspec', tdir)
    monkeykas.chdir(tdir)
    with mercurial(tmpdir, 'evolve', branch='stable'):
        with pytest.raises(RepoRefError):
            kas.kas(['checkout', 'test9.yml'])

    with pytest.raises(RepoRefError):
        kas.kas(['checkout', 'test10.yml'])

    with pytest.raises(RepoRefError):
        kas.kas(['checkout', 'test11.yml'])


@pytest.mark.online
def test_commit_expand(monkeykas, tmpdir, capsys):
    """
        Test if an abbreviated commit hash is expanded to the full hash.
    """
    tdir = str(tmpdir / 'test_commit_expand')
    shutil.copytree('tests/test_refspec', tdir)
    monkeykas.chdir(tdir)
    kas.kas(['dump', '--resolve-refs', 'test12.yml'])
    rawspec = yaml.safe_load(capsys.readouterr().out)
    assert rawspec['repos']['kas']['commit'] == \
        'abd109469d17b7ff4d958b5aa5ab5f5511cc4d43'



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This module provides infrastructure to generate provenance attestation
    of the build process.
"""

import os
import distro
import logging
import hashlib
import base64
import sys
from enum import Enum
from urllib.parse import urlparse
from pathlib import Path
from datetime import datetime, timezone
from kas import __version__ as KASVERSION


SLSA_PROVENANCE_TYPE = 'https://slsa.dev/provenance/v1'
KAS_BUILDER_ID = 'https://github.com/siemens/kas'
KAS_BUILD_TYPE = 'https://kas.readthedocs.io/en' \
                 f'/{KASVERSION}/userguide/project-configuration.html'
INTOTO_STATEMENT_TYPE = 'https://in-toto.io/Statement/v1'


def date_to_rfc3339(dt):
    return dt.astimezone(timezone.utc) \
             .strftime('%Y-%m-%dT%H:%M:%S.%fZ')


def file_digest_slow(f, algorithm):
    """
        Implementation of hashlib.file_digest for python < 3.11
    """
    hash = hashlib.new(algorithm)
    while True:
        data = f.read()
        if not data:
            break
        hash.update(data)
    return hash


class Provenance:
    """
        Create a build attestation predicate in SLSA provenance format
        for a kas build.
    """
    class Mode(Enum):
        MIN = 1
        MAX = 10

    def __init__(self, ctx, t_started, t_finished, mode=Mode.MIN):
        self._ctx = ctx
        self._t_started = t_started
        self._t_finished = t_finished
        self._mode = mode

    @staticmethod
    def _url_with_protocol(url):
        if url.startswith('git@'):
            return f'ssh://{url}'
        if url.startswith('ssh://'):
            return f'{url}'
        if url.startswith('http://') or url.startswith('https://'):
            return f'{url}'

    @staticmethod
    def _strip_credentials(url):
        """
            Returns the url with all credentials removed (best effort)
        """
        if not url.startswith('http://') and not url.startswith('https://'):
            return url
        parsed = urlparse(url)
        if parsed.username and parsed.password:
            url = url.replace(f'{parsed.username}:{parsed.password}@', '')
        return url

    @staticmethod
    def _get_filetype(f: Path):
        if f.suffix == '.json':
            return 'json'
        return 'yaml'

    def _make_relative_path(self, path: Path):
        top_repo_path = Path(self._ctx.config.handler.get_top_repo_path())
        workdir = Path(self._ctx.kas_work_dir)

        if path.is_relative_to(workdir):
            return path.relative_to(workdir)
        else:
            return path.relative_to(top_repo_path)

    def type_(self):
        return SLSA_PROVENANCE_TYPE

    def as_dict(self):
        res_deps = []
        tracked_repos = []
        for r in self._ctx.config.get_repos():
            if r.operations_disabled:
                if not r.url or not r.revision:
                    continue
            digest = {f'{r.get_type()}Commit': r.revision}
            annotations = {
                'dirty': r.dirty,
                'layers': [str(layer.path.relative_to(r.path))
                           for layer in r.layers]
            }
            cleanurl = self._strip_credentials(r.url)
            dep = {
                'name': r.name,
                'uri': f'{r.get_type()}+{self._url_with_protocol(cleanurl)}',
                'digest': digest,
                'annotations': annotations
            }
            res_deps.append(dep)
            tracked_repos.append(r)

        # (abspath, relpath)
        config_files = [(Path(c), self._make_relative_path(Path(c)))
                        for c in self._ctx.config.filenames]
        for ca, cr in config_files:
            if any([r.contains_path(cr) for r in tracked_repos]):
                logging.debug(f'Config file {cr} is tracked')
                continue

            with open(ca, 'rb') as f:
                content = f.read()
            rd = {
                'name': str(cr),
                'content': base64.b64encode(content).decode('utf-8'),
                'mediaType': f'application/vnd.kas+{self._get_filetype(ca)}'
            }
            res_deps.append(rd)

        bd = {
            'buildType': KAS_BUILD_TYPE,
            'externalParameters': {
                'command': self._ctx.args.cmd,
                'config': [str(c) for _, c in config_files],
                'target': self._ctx.args.target,
                'task': self._ctx.args.task,
                'extra_bitbake_args': self._ctx.args.extra_bitbake_args,
            },
            'internalParameters': {},
            'resolvedDependencies': res_deps
        }
        if self._mode == self.Mode.MAX:
            bd['internalParameters']['env'] = \
                self._ctx.config.get_environment()
        b_versions = {
            'kas': KASVERSION,
            'distro.name': distro.id(),
            'distro.version': distro.version()
        }

        rd = {
            'builder':
            {
                'id': KAS_BUILDER_ID,
                'version': b_versions
            },
            'metadata': {
                'invocationId': os.environ.get('CI_JOB_URL', ''),
                'startedOn': date_to_rfc3339(self._t_started),
                'finishedOn': date_to_rfc3339(self._t_finished),
            },
            'byproducts': []
        }
        p = {
            'buildDefinition': bd,
            'runDetails': rd
        }
        return p


class Statement:
    """
        Create a statement in in-toto format for a kas build.
    """
    def __init__(self, predicate, ctx, t_started, t_finished):
        self._predicate = predicate
        self._ctx = ctx
        self._t_started = t_started
        self._t_finished = t_finished

    def _check_artifact_timestamp(self, name, path):
        """
            Warn if artifact timestamp is not within the build range.
        """
        logging.debug(f'Found artifact {name}:{path} in build dir')
        fullpath = Path(self._ctx.build_dir) / path
        mtime = datetime.fromtimestamp(fullpath.stat().st_mtime)
        if mtime < self._t_started or mtime > self._t_finished:
            logging.warning(
                f'Artifact {name}:{path.name} mtime {mtime.strftime("%c")}'
                f' not in build range '
                f'[{self._t_started.strftime("%c")} - '
                f'{self._t_finished.strftime("%c")}]')

    def as_dict(self):
        pt = self._predicate.type_()
        pp = self._predicate.as_dict()
        subjects = []
        for n, s in self._ctx.config.get_artifacts(missing_ok=False):
            self._check_artifact_timestamp(n, s)
            fullpath = Path(self._ctx.build_dir) / s
            with open(fullpath, "rb") as f:
                if sys.version_info < (3, 11):
                    digest = file_digest_slow(f, "sha256")
                else:
                    digest = hashlib.file_digest(f, "sha256")
            rd = {
                'name': s.name,
                'digest': {'sha256': digest.hexdigest()}
            }
            subjects.append(rd)
        if len(subjects) == 0:
            logging.warning('Attestation does not contain any artifacts.')
        st = {
            '_type': INTOTO_STATEMENT_TYPE,
            'subject': subjects,
            'predicateType': pt,
            'predicate': pp
        }
        return st



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens, 2025
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This module provide the infrastructure to setup and work with local
    GnuPG keyrings.
"""

import logging
import subprocess
from pathlib import Path
from git.config import GitConfigParser
from kas.kasusererror import KasUserError, MissingModuleError

try:
    import gnupg
    HAVE_GNUPG = True
except ImportError:
    HAVE_GNUPG = False


class KeyImportError(KasUserError):
    """
    Raised when a key could not be imported.
    """
    def __init__(self, name, message):
        super().__init__(f'Could not import key "{name}": {message}')


class KeyHandler:
    @property
    def env(self):
        return {}

    def prepare_validation(self, repo):
        """
        Configure the repository for signature validation (e.g.
        define set of allowed keys). Currently only supported for
        git repos.
        """
        if repo.get_type() != 'git':
            raise KasUserError('Only git repositories are supported')

    def validate_allowed_signer(self, repo, output):
        """
        Process the output of a signature validation and validate
        it against the list of allowed signers. Must only be called
        if the validation by the VCS was completed successfully.
        """
        assert repo.get_type() == 'git'

    def get_key_repr(self, keyname):
        """
        Returns a human readable representation of the key. Derived
        classes may override this to provide a more meaningful
        representation of the key.
        """
        return keyname


class GPGKeyHandler(KeyHandler):
    def __init__(self, gnupghome, signers, confinst):
        if not HAVE_GNUPG:
            raise MissingModuleError('python-gnupg', 'signature verification')

        if logging.getLogger().level <= logging.DEBUG:
            logging.getLogger('gnupg').setLevel(logging.INFO)

        self.gpg = gnupg.GPG(gnupghome=str(gnupghome))
        self.fingerprints = {}
        self._reset_trust()
        self._import_keys(signers, confinst)

    def _reset_trust(self):
        all_keys = self.gpg.list_keys()
        for key in all_keys:
            fingerprint = key['fingerprint']
            logging.debug(f'Resetting trust level for key "{fingerprint}"')
            self.gpg.trust_keys(fingerprint, 'TRUST_NEVER')

    def _import_keys(self, signers, confinst):
        for name, loc in signers.items():
            fingerprint = loc.get('fingerprint')
            keyserver = loc.get('gpg_keyserver')
            if 'path' in loc:
                if 'repo' not in loc:
                    raise KasUserError('path must be used together with repo')
                repo = confinst.get_repo(loc['repo'])
                keyfile = Path(repo.path) / Path(loc['path'])
                import_result = self.gpg.import_keys_file(keyfile)
            elif keyserver:
                if not fingerprint:
                    raise KasUserError('"gpg_keyserver" must be used together '
                                       'with "fingerprint"')
                import_result = self.gpg.recv_keys(keyserver,
                                                   fingerprint)
            if import_result.count == 0:
                raise KeyImportError(name, 'No keys imported')
            if import_result.count > 1:
                raise KeyImportError(name,
                                     'Multiple keys imported')
            # an import result can also have expired keys which are not
            # imported but show up in the results array. Ignore them.
            actual_fp = [k for k in import_result.results
                         if k['fingerprint']][0]['fingerprint']

            if fingerprint and actual_fp != fingerprint:
                raise KeyImportError(name,
                                     'Fingerprints do not match: '
                                     f'Key has "{actual_fp}". '
                                     f'Expected "{fingerprint}".')
            # we operate on a kas-local keystore, so we can trust the key
            self.gpg.trust_keys(actual_fp, 'TRUST_ULTIMATE')
            self.fingerprints[name] = actual_fp
            logging.debug(f'Imported key "{name}" with fingerprint '
                          f'"{actual_fp}"')

    def _fingerprint(self, keyname):
        fingerprint = self.fingerprints.get(keyname)
        if not fingerprint:
            raise KasUserError(f'Key "{keyname}" not found')
        return self.gpg.list_keys(keys=fingerprint)[0]['fingerprint']

    def _keyid(self, fingerprint):
        keyname = [k for k, v in self.fingerprints.items()
                   if v == fingerprint]
        if not keyname:
            return None
        return keyname[0]

    @property
    def env(self):
        return {'GNUPGHOME': str(self.gpg.gnupghome)}

    def validate_allowed_signer(self, repo, output):
        super().validate_allowed_signer(repo, output)
        allowed_fps = [self._fingerprint(x) for x in repo.allowed_signers]
        validsigs = [x for x in output.split('\n') if 'VALIDSIG' in x]
        sigs = [x.split()[-1] for x in validsigs]
        if not sigs:
            return (False, None)
        for sig in sigs:
            if sig in allowed_fps:
                return (True, self._keyid(sig))
        return (False, self._keyid(sigs[0]))

    def get_key_repr(self, keyname):
        fingerprint = self._fingerprint(keyname)
        uids = self.gpg.list_keys(keys=fingerprint)[0]['uids']
        fp_formatted = ' '.join([(fingerprint[i:i + 4])
                                for i in range(0, len(fingerprint), 4)])
        uidstr = ' aka '.join([f'"{uid}"' for uid in uids])
        return f'Fingerprint {fp_formatted} from {uidstr}'


class SSHKeyHandler(KeyHandler):
    def __init__(self, workdir, signers, confinst):
        self.workdir = workdir
        self.signers = {}

        for name, loc in signers.items():
            if 'repo' not in loc:
                raise KasUserError('For SSH keys, a repo must be specified')
            if 'path' not in loc:
                raise KasUserError('For SSH keys, a path must be specified')
            repo = confinst.get_repo(loc['repo'])
            pubfile = Path(repo.path) / Path(loc['path'])
            keydata = subprocess.check_output(['ssh-keygen', '-lf',
                                               pubfile.absolute()])
            keyparts = keydata.decode('utf-8').split()
            size, fp = keyparts[0:2]
            comment = ' '.join(keyparts[2:-1])
            rawtype = keyparts[-1].replace('(', '').replace(')', '')
            keytype = self._key_name_from_sn(rawtype, size)
            if 'fingerprint' in loc and loc['fingerprint'] != fp:
                raise KeyImportError(name,
                                     'Fingerprints do not match: '
                                     f'Key has "{fp}". '
                                     f'Expected "{loc["fingerprint"]}".')
            rawkey = subprocess.check_output(['ssh-keygen', '-mRFC4716', '-ef',
                                              pubfile])
            key = self._key_from_rfc4716(rawkey.decode('utf-8'))
            self.signers[name] = {
                'type': keytype.lower(),
                'key': key,
                'comment': comment,
                'fingerprint': fp,
                'size': size
            }

    @staticmethod
    def _key_name_from_sn(name, size):
        """
        Convert the key type and size from the output of ssh-keygen -lf
        to the ssh key type string.
        """
        # RFC 4253 section 6.6
        if name == 'DSA':
            return 'ssh-dss'
        if name == 'RSA':
            return 'ssh-rsa'
        # RFC 5656 section 6.2
        if name == 'ECDSA' and int(size) in [256, 384, 521]:
            return f'ecdsa-sha2-nistp{size}'
        # RFC 8709 section 6
        if name == 'ED25519':
            return 'ssh-ed25519'
        raise KasUserError(f'Unsupported key type "{name}" with size "{size}"')

    @staticmethod
    def _key_from_rfc4716(data):
        parts = []
        for line in data.splitlines():
            if not line.startswith('----') and not line.startswith('Comment:'):
                parts.append(line)
        return ''.join(parts)

    def _keyid(self, fingerprint):
        for k, v in self.signers.items():
            if v['fingerprint'] == fingerprint:
                return k
        return None

    def prepare_validation(self, repo):
        super().prepare_validation(repo)
        trustpath = self.workdir / repo.name
        trustpath.mkdir(exist_ok=True)
        allowedSignersFile = trustpath / 'allowedSigners'

        with open(allowedSignersFile, 'w') as f:
            for k in repo.allowed_signers:
                signer = self.signers.get(k)
                f.write(f'"{signer["comment"]}" namespaces="git" '
                        f'{signer["type"]} {signer["key"]}\n')

        gitconfig = Path(repo.path) / '.git/config'
        with GitConfigParser(gitconfig, read_only=False) as config:
            config.add_section('gpg "ssh"')
            config['gpg "ssh"']['allowedSignersFile'] = str(allowedSignersFile)
            config.write()

    def validate_allowed_signer(self, repo, output):
        super().validate_allowed_signer(repo, output)
        fp = output.split()[-1]
        return (True, self._keyid(fp))

    def get_key_repr(self, keyname):
        signer = self.signers.get(keyname)
        return f'Fingerprint {signer["fingerprint"]} ({signer["type"]}) ' \
               f'from "{signer["comment"]}"'



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import shutil
import snack
import pytest
from kas import kas


@pytest.fixture(autouse=True)
def patch_kas(monkeykas):
    INPUTS = iter([' ', None, ' ', None])
    ACTIONS = iter([None, 'build', None, 'build'])
    SELECTIONS = iter([0, 3])

    def mock_runOnce(unused1):
        return next(INPUTS)

    def mock_buttonPressed(unused1, unused2):
        return next(ACTIONS)

    def mock_current(unused1):
        return next(SELECTIONS)

    monkeykas.setattr(snack.GridFormHelp, 'runOnce', mock_runOnce)
    monkeykas.setattr(snack.ButtonBar, 'buttonPressed', mock_buttonPressed)
    monkeykas.setattr(snack.Listbox, 'current', mock_current)


def file_contains(filename, expected):
    with open(filename) as file:
        for line in file.readlines():
            if line == expected:
                return True
    return False


def check_bitbake_options(expected, build_dir):
    with open(build_dir / 'bitbake.options') as file:
        return file.readline() == expected


def test_menu(monkeykas, tmpdir):
    tdir = str(tmpdir / 'test_menu')
    shutil.copytree('tests/test_menu', tdir)
    monkeykas.chdir(tdir)

    kas_wd = monkeykas.get_kwd()
    kas_bd = monkeykas.get_kbd()
    local_conf = kas_bd / 'conf/local.conf'
    # select opt1 & build
    kas.kas(['menu'])
    assert file_contains(local_conf, 'OPT1 = "1"\n')
    assert file_contains(kas_wd / '.config.yaml',
                         'build_system: openembedded\n')
    assert check_bitbake_options('-c build target1\n', kas_bd)

    # rebuild test
    kas.kas(['build'])
    assert file_contains(local_conf, 'OPT1 = "1"\n')
    assert check_bitbake_options('-c build target1\n', kas_bd)

    # select alternative target & build
    kas.kas(['menu'])
    assert file_contains(local_conf, 'OPT1 = "1"\n')
    assert check_bitbake_options('-c build target2\n', kas_bd)


@pytest.mark.dirsfromenv
def test_menu_inc_workdir(monkeykas, tmpdir):
    tdir = str(tmpdir / 'test_menu_inc')
    kas_workdir = str(tmpdir / 'test_menu_inc' / 'out')
    shutil.copytree('tests/test_menu', tdir)
    monkeykas.chdir(tdir)
    os.mkdir(kas_workdir)
    monkeykas.setenv('KAS_WORK_DIR', kas_workdir)
    kas.kas(['menu'])
    default_config = os.path.join(kas_workdir, '.config.yaml')
    assert os.path.exists(default_config)
    # check if purge removes all files (including the generated .config.yml)
    kas.kas(['purge'])
    assert not os.path.exists(default_config)


@pytest.mark.dirsfromenv
def test_menu_implicit_workdir(monkeykas, tmpdir):
    tdir = str(tmpdir / 'test_menu_iwd')
    kas_workdir = str(tmpdir / 'test_menu_iwd_out')
    shutil.copytree('tests/test_menu', tdir)
    os.mkdir(kas_workdir)
    monkeykas.chdir(kas_workdir)
    kas.kas(['menu', tdir + '/Kconfig'])



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Parts of this were based on kconfiglib, examples/menuconfig_example.py
#
# Copyright (c) 2011-2019, Ulf Magnusson <ulfalizer@gmail.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
"""
    This plugin implements the ``kas menu`` command.

    When this command is executed, kas will open a configuration menu as
    described by a Kconfig file. It processes any pre-existing configuration
    file with saved settings, stores the final selections and invokes the build
    plugin if requested by the user.

    To make use of this plugin, a ``Kconfig`` file has to be provided. The
    menu can define these types of configuration variables that the plugin
    will translate into a kas configuration:

     - kas configuration files that will be included when building the
       generated configuration. Those are picked up from Kconfig string
       variables that have the name prefix ``KAS_INCLUDE_``.

     - bitbake targets that shall be built via the generated configuration.
       Those are picked up from Kconfig string variables that have the name
       prefix ``KAS_TARGET_``.

     - The ``build_system`` that will used. The static Kconfig string variable
       ``KAS_BUILD_SYSTEM`` defines this value which must be ``openembedded``,
       ``oe`` or ``isar`` is set.

     - bitbake configuration variables that will be added to the
       ``local_conf_header`` section of the generated configuration. All other
       active Kconfig string, integer or hex variables are treated as such.

    See https://www.kernel.org/doc/html/latest/kbuild/kconfig-language.html
    for a complete documentation of the Kconfig language.

    The menu plugin writes the selected configuration to a ``.config.yaml``
    file in the kas work directory and also reads previous selection from such
    a file if it exists. The ``.config.yaml`` both contains the selected
    configuration in the ``menu_configuration`` key and also the effective
    settings that can be used to invoke ``kas build`` or other kas commands.
"""

import logging
import os
import pprint
import yaml
from kas import __version__, __file_version__
from kas.context import create_global_context
from kas.config import CONFIG_YAML_FILE
from kas.repos import Repo
from kas.includehandler import ConfigFile, \
    SOURCE_DIR_OVERRIDE_KEY, SOURCE_DIR_HOST_OVERRIDE_KEY
from kas.plugins.build import Build
from kas.kasusererror import KasUserError, MissingModuleError

try:
    from kconfiglib import Kconfig, Symbol, Choice, KconfigError, \
        expr_value, TYPE_TO_STR, MENU, COMMENT, STRING, BOOL, INT, HEX, UNKNOWN
    HAVE_KCONFIGLIB = True
except ImportError:
    HAVE_KCONFIGLIB = False  # will be reported in run()

try:
    from snack import SnackScreen, EntryWindow, ButtonChoiceWindow, \
        ButtonBar, Listbox, GridFormHelp
    HAVE_NEWT = True
except ImportError:
    HAVE_NEWT = False  # will be reported in run()

__license__ = 'MIT'
__copyright__ = \
    'Copyright (c) 2011-2019, Ulf Magnusson <ulfalizer@gmail.com>\n' \
    'Copyright (c) Siemens AG, 2021-2023'

SOURCE_DIR_HOST_ENV_KEY = '_KAS_REPO_DIR_HOST'


class VariableTypeError(KasUserError):
    pass


class KConfigLoadError(KasUserError):
    """
    The KConfig file could not be found or is invalid
    """
    pass


def check_sym_is_string(sym):
    if sym.type != STRING:
        raise VariableTypeError(f'Variable {sym.name} must be of string type')


def str_representer(dumper, data):
    style = '|' if len(data.splitlines()) > 1 else None
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style=style)


class Args:
    pass


class Menu:
    """
        This class implements the menu plugin for kas.
    """

    name = 'menu'
    helpmsg = (
        'Provides a configuration menu and triggers the build of the choices.'
    )

    @classmethod
    def setup_parser(cls, parser):
        parser.add_argument('kconfig',
                            help='Kconfig file',
                            nargs='?', default='Kconfig')

    def load_config(self, filename):
        try:
            config = ConfigFile.load(filename)
            self.orig_config = config.config
        except FileNotFoundError:
            self.orig_config = {}
            return

        menu_configuration = self.orig_config.get('menu_configuration', {})
        for symname in menu_configuration:
            sym = self.kconf.syms.get(symname)
            if not sym:
                logging.warning(
                    'Ignoring unknown configuration variable %s in %s',
                    symname, filename)
                continue
            symvalue = menu_configuration[symname]
            if sym.type == BOOL:
                sym.set_value('y' if symvalue else 'n')
            elif sym.type == INT:
                sym.set_value(str(symvalue))
            elif sym.type == HEX:
                sym.set_value(str(hex(symvalue)))
            else:  # string
                sym.set_value(symvalue)

    def save_config(self, filename, top_repo_dir):
        kas_includes = []
        kas_targets = []
        kas_build_system = None
        kas_vars = {}
        menu_configuration = {}

        for symname in self.kconf.syms:
            if symname == 'MODULES':
                continue

            sym = self.kconf.syms[symname]
            symvalue = sym.str_value

            if expr_value(sym.direct_dep) < 2:
                continue

            if sym.visibility == 2:
                if sym.type == BOOL:
                    menu_configuration[symname] = symvalue == 'y'
                elif sym.type == STRING:
                    menu_configuration[symname] = symvalue
                elif sym.type == INT:
                    menu_configuration[symname] = int(symvalue)
                elif sym.type == HEX:
                    menu_configuration[symname] = int(symvalue, 16)
                else:
                    raise VariableTypeError(
                        'Configuration variable {symname} uses unsupported '
                        'type')

            if symname.startswith('KAS_INCLUDE_'):
                check_sym_is_string(sym)
                if symvalue != '':
                    kas_includes.append(symvalue)
            elif symname.startswith('KAS_TARGET_'):
                check_sym_is_string(sym)
                if symvalue != '':
                    kas_targets.append(symvalue)
            elif symname == 'KAS_BUILD_SYSTEM':
                check_sym_is_string(sym)
                if symvalue != '':
                    kas_build_system = symvalue
            elif sym.type in (STRING, INT, HEX):
                kas_vars[symname] = symvalue

        config = {
            'header': {
                'version': __file_version__,
                'includes': kas_includes
            },
            'menu_configuration': menu_configuration,
            SOURCE_DIR_OVERRIDE_KEY: top_repo_dir
        }

        if SOURCE_DIR_HOST_ENV_KEY in os.environ:
            config[SOURCE_DIR_HOST_OVERRIDE_KEY] = \
                os.environ[SOURCE_DIR_HOST_ENV_KEY]
        if kas_build_system:
            config['build_system'] = kas_build_system
        if len(kas_targets) > 0:
            config['target'] = kas_targets
        if len(kas_vars) > 0:
            config['local_conf_header'] = {
                '__menu_config_vars': '\n'.join([
                    f'{key} = "{value}"'
                    for key, value in kas_vars.items()
                ])
            }

        logging.debug('Menu configuration:\n%s', pprint.pformat(config))

        if config != self.orig_config:
            logging.info('Saving configuration as %s', filename)

            # format multi-line strings more nicely
            yaml.add_representer(str, str_representer)

            try:
                os.rename(filename, filename + '.old')
            except FileNotFoundError:
                pass

            with open(filename, 'w') as config_file:
                config_file.write(
                    '#\n'
                    f'# Automatically generated by kas {__version__}\n'
                    '#\n')
                yaml.dump(config, config_file)

    def dump_kconf_warnings(self):
        if len(self.kconf.warnings) > 0:
            logging.warning("\n".join(self.kconf.warnings))
            self.kconf.warnings = []

    def run(self, args):
        if not HAVE_KCONFIGLIB:
            raise MissingModuleError('python3-kconfiglib', 'Menu plugin')
        if not HAVE_NEWT:
            raise MissingModuleError('python3-newt', 'Menu plugin')

        ctx = create_global_context(args)

        kconfig_file = os.path.abspath(args.kconfig)
        try:
            self.kconf = Kconfig(kconfig_file, warn_to_stderr=False)
        except (KconfigError, FileNotFoundError) as err:
            raise KConfigLoadError(str(err))

        top_repo_path = Repo.get_root_path(os.path.dirname(kconfig_file))
        config_filename = os.path.join(ctx.kas_work_dir, CONFIG_YAML_FILE)

        self.load_config(config_filename)
        self.dump_kconf_warnings()

        menu = Menuconfig(self.kconf)
        action = menu.show()

        if action == 'exit':
            return

        self.save_config(config_filename, top_repo_path)
        self.dump_kconf_warnings()

        if action == 'build':
            logging.debug('Starting build')

            build_args = Args()
            build_args.config = None
            build_args.target = None
            build_args.task = None
            build_args.extra_bitbake_args = []
            build_args.skip = None
            build_args.provenance = False

            Build().run(build_args)


class Menuconfig():
    def __init__(self, kconf):
        self.kconf = kconf
        self.screen = None

    @staticmethod
    def value_str(sym):
        """
        Returns the value part ("[*]", "(foo)" etc.) of a menu entry.
        """
        if sym.type in (STRING, INT, HEX):
            return f"({sym.str_value})"

        # BOOL (TRISTATE not supported)

        # The choice mode is an upper bound on the visibility of choice
        # symbols, so we can check the choice symbols' own visibility to see
        # if the choice is in y mode
        if sym.choice and sym.visibility == 2:
            return "(*)" if sym.choice.selection is sym else "( )"

        tri_val_str = (" ", None, "*")[sym.tri_value]

        if len(sym.assignable) == 1:
            # Pinned to a single value
            return f"-{tri_val_str}-"

        if sym.type == BOOL:
            return f"[{tri_val_str}]"

        raise RuntimeError()

    @staticmethod
    def node_str(node, indent):
        """
        Returns the complete menu entry text for a menu node, or "" for
        invisible menu nodes. Invisible menu nodes are those that lack a prompt
        or that do not have a satisfied prompt condition.

        Example return value: "[*] Bool symbol (BOOL)"

        The symbol name is printed in parentheses to the right of the prompt.
        This is so that symbols can easily be referred to in the configuration
        interface.
        """
        if not node.prompt:
            return ""

        # Even for menu nodes for symbols and choices, it's wrong to check
        # Symbol.visibility / Choice.visibility here. The reason is that a
        # symbol (and a choice, in theory) can be defined in multiple
        # locations, giving it multiple menu nodes, which do not necessarily
        # all have the same prompt visibility. Symbol.visibility /
        # Choice.visibility is calculated as the OR of the visibility of all
        # the prompts.
        prompt, prompt_cond = node.prompt
        if not expr_value(prompt_cond):
            return ""

        if node.item == MENU:
            return f"    {indent * ' '}{prompt}  --->"

        if type(node.item) is Choice:
            return f"    {indent * ' '}{prompt}"

        if node.item == COMMENT:
            return f"    {indent * ' '}*** {prompt} ***"

        # Symbol
        sym = node.item

        if sym.type == UNKNOWN:
            return ""

        # {:3} sets the field width to three. Gives nice alignment for empty
        # string values.
        res = f"{Menuconfig.value_str(sym):3} {indent * ' '}{prompt}"

        # Append a sub-menu arrow if menuconfig and enabled
        if node.is_menuconfig:
            res += f"  ---{'>' if sym.tri_value > 0 else '-'}"

        return res

    @staticmethod
    def menu_node_strings(node, indent):
        items = []

        while node:
            string = Menuconfig.node_str(node, indent)
            if string:
                items.append((string, node))

            if (node.list and node.item != MENU
                    and (type(node.item) is Choice or not node.is_menuconfig)):
                items.extend(Menuconfig.menu_node_strings(node.list,
                                                          indent + 2))

            node = node.next

        return items

    def show_menu(self, title, top_node, is_submenu=False):
        selection = 0

        while True:
            items = Menuconfig.menu_node_strings(top_node, 0)

            height = len(items)
            scroll = 0
            if height > self.screen.height - 13:
                height = self.screen.height - 13
                scroll = 1

            buttons = [
                ('Build', 'build', 'B'),
                ('Save & Exit', 'save', 'S'),
                (' Exit ', 'exit', 'E'),
                (' Help ', 'help', 'h')
            ]
            if is_submenu:
                buttons.insert(0, (' Return ', 'return', 'ESC'))
            buttonbar = ButtonBar(self.screen, buttons)
            if not is_submenu:
                buttonbar.hotkeys['ESC'] = 'exit'
            listbox = Listbox(height, scroll=scroll, returnExit=1)
            count = 0
            for string, _ in items:
                listbox.append(string, count)
                if (selection == count):
                    listbox.setCurrent(count)
                count = count + 1

            grid = GridFormHelp(self.screen, title, None, 1, 2)
            grid.add(listbox, 0, 0, padding=(0, 0, 0, 1))
            grid.add(buttonbar, 0, 1, growx=1)
            grid.addHotKey(' ')

            rc = grid.runOnce()

            action = buttonbar.buttonPressed(rc)
            if action and action != 'help':
                return action

            if count == 0:
                continue

            selection = listbox.current()
            _, selected_node = items[selection]
            sym = selected_node.item

            if action == 'help':
                prompt, _ = selected_node.prompt
                if hasattr(selected_node, 'help') and selected_node.help:
                    help = selected_node.help
                else:
                    help = 'No help available.'
                ButtonChoiceWindow(
                    screen=self.screen,
                    title=f"Help on '{prompt}'",
                    text=help,
                    width=60,
                    buttons=['  Ok  '])
                continue

            show_submenu = False

            if type(sym) is Symbol:
                if rc == ' ':
                    if sym.type == BOOL:
                        sym.set_value('n' if sym.tri_value > 0 else 'y')
                else:
                    if selected_node.is_menuconfig:
                        show_submenu = True
                    elif sym.type in (STRING, INT, HEX):
                        action, values = EntryWindow(
                            screen=self.screen,
                            title=sym.name,
                            text=f'Enter a {TYPE_TO_STR[sym.type]} value:',
                            prompts=[('', sym.str_value)],
                            buttons=[('  Ok  ', 'Ok'), ('Cancel', '', 'ESC')])
                        if action == 'Ok':
                            self.kconf.warnings = []
                            val = values[0]
                            if sym.type == HEX and not val.startswith('0x'):
                                val = '0x' + val
                            sym.set_value(val)
                            # only fetching triggers range check - how ugly...
                            sym.str_value
                            if len(self.kconf.warnings) > 0:
                                ButtonChoiceWindow(
                                    screen=self.screen,
                                    title="Invalid entry",
                                    text="\n".join(self.kconf.warnings),
                                    width=60,
                                    buttons=['  Ok  '])
                                self.kconf.warnings = []
            elif selected_node.is_menuconfig and type(sym) is not Choice:
                show_submenu = True

            if show_submenu:
                submenu_title, _ = selected_node.prompt
                action = self.show_menu(submenu_title,
                                        selected_node.list,
                                        is_submenu=True)
                if action != 'return':
                    return action

    def show(self):
        self.screen = SnackScreen()

        action = self.show_menu(self.kconf.mainmenu_text,
                                self.kconf.top_node.list)

        self.screen.finish()
        return action


__KAS_PLUGINS__ = [Menu]



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2017-2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This plugin implements the ``kas build`` command.

    When this command is executed, kas will checkout repositories, setup the
    build environment and then invoke bitbake to build the targets selected
    in the chosen config file.

    When running with ``--provenance <true|mode=...>`` kas will generate an
    provenance attestation for the build. The attestation will be stored in
    ``attestation/kas-build.provenance.json`` in the build directory.
    For details about provenance, see the build attestation chapter.

    .. note::
        In provenance mode, the command returns with a non-zero exit
        code in case no artifact is found for at least one entry.
"""

import logging
import subprocess
import sys
import json
import asyncio
from pathlib import Path
from datetime import datetime
from kas.context import create_global_context
from kas.config import Config
from kas.libkas import find_program, run_cmd_async
from kas.libkas import setup_parser_keep_config_unchanged_arg
from kas.libcmds import Macro, Command
from kas.libkas import setup_parser_common_args, setup_parser_config_arg
from kas.kasusererror import CommandExecError
from kas.attestation import Provenance, Statement


__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens AG, 2017-2024'


class Build:
    """
        This class implements the build plugin for kas.
    """

    name = 'build'
    helpmsg = (
        'Checks out all necessary repositories and builds using bitbake as '
        'specified in the configuration file.'
    )

    @classmethod
    def setup_parser(cls, parser):
        """
            Setup the argument parser for the build plugin
        """

        setup_parser_common_args(parser)
        setup_parser_config_arg(parser)
        setup_parser_keep_config_unchanged_arg(parser)
        parser.add_argument('extra_bitbake_args',
                            nargs='*',
                            help='Extra arguments to pass to bitbake '
                                 '(typically requires separation via \'--\')')
        parser.add_argument('--target',
                            action='append',
                            help='Select target to build')
        parser.add_argument('-c', '--cmd', '--task', dest='task',
                            help='Select which task should be executed')
        parser.add_argument('--provenance',
                            choices=['true', 'mode=min', 'mode=max'],
                            help='Enable provenance attestation generation')

    def run(self, args):
        """
            Executes the build command of the kas plugin.
        """

        if args.config and args.config.startswith('-'):
            args.extra_bitbake_args.insert(0, args.config)
            args.config = None

        ctx = create_global_context(args)
        ctx.config = Config(ctx, args.config, args.target, args.task)

        macro = Macro()
        macro.add(BuildCommand(args.extra_bitbake_args))
        macro.run(ctx, args.skip)


class BuildCommand(Command):
    """
        Implements the bitbake build step.
    """

    def __init__(self, extra_bitbake_args):
        super().__init__()
        self.extra_bitbake_args = extra_bitbake_args

    def __str__(self):
        return 'build'

    def _generate_attestation(self, ctx,
                              time_started, time_finished,
                              mode):
        """
            Generate the provenance attestation for the build.
        """
        predicate = Provenance(ctx, time_started, time_finished,
                               mode)
        stmt = Statement(predicate, ctx, time_started, time_finished).as_dict()
        att_dir = Path(ctx.build_dir) / 'attestation'
        att_dir.mkdir(parents=True, exist_ok=True)
        with open(att_dir / 'kas-build.provenance.json', 'w') as f:
            f.write(json.dumps(stmt, indent=4))
            f.write('\n')

    def execute(self, ctx):
        """
            Executes the bitbake build command.
        """
        # Start bitbake build of image
        bitbake = find_program(ctx.environ['PATH'], 'bitbake')
        cmd = [bitbake, '-c', ctx.config.get_bitbake_task()] \
            + self.extra_bitbake_args + ctx.config.get_bitbake_targets()

        time_started = datetime.now()
        if sys.stdout.isatty():
            logging.info('%s$ %s', ctx.build_dir, ' '.join(cmd))
            ret = subprocess.call(cmd, env=ctx.environ, cwd=ctx.build_dir)
            if ret != 0:
                raise CommandExecError(cmd, ret)
        else:
            loop = asyncio.get_event_loop()
            loop.run_until_complete(run_cmd_async(cmd, cwd=ctx.build_dir,
                                                  liveupdate=True))
        time_finished = datetime.now()

        if ctx.args.provenance:
            mode = Provenance.Mode.MAX if ctx.args.provenance == 'mode=max' \
                else Provenance.Mode.MIN
            self._generate_attestation(ctx, time_started, time_finished, mode)


__KAS_PLUGINS__ = [Build]



#!/usr/bin/env python3
# takes a reverse-sorted, line separated list and
# returns the first element that is equal or smaller
# than the first argument

import sys

for line in sys.stdin:
    if line.rstrip() <= sys.argv[1]:
        print(line.rstrip())
        break
